#include	"u.h"
#include	"../port/lib.h"
#include	"mem.h"
#include	"dat.h"
#include	"fns.h"
#include	"io.h"
#include	"ureg.h"
#include	"../port/error.h"

/* we use 64 HZ for rtc, and timer 3 (32-bits) for OS timer */
enum {
	Never	=	0xffffffff,
};

/* the RTC wil interrupt at 64 HZ. 
 * the OS timer is the 32-bit timer 3, running at 508 KHZ.
 */

OSTimerRegs *timerregs = (OSTimerRegs*)OSTIMERREGS;
static int clockinited;

static void	clockintr(Ureg*, void*);
static void	rtcintr(Ureg*, void*);
static uvlong	when;	/* scheduled time of next interrupt */

long	timeradjust;

enum
{
	Minfreq = ClockFreq/HZ,		/* At least one interrupt per HZ (50 ms) */
	Maxfreq = ClockFreq/10000,	/* At most one interrupt every 100 µs */
};

/*
ulong
clockpower(int on)
{
	static ulong savedtime;

	if (on){
		timerregs->ossr |= 1<<0;
		timerregs->oier = 1<<0;
		timerregs->osmr[0] = timerregs->oscr + Minfreq;
		if (rtcregs->rttr == 0){
			rtcregs->rttr = 0x8000; // nominal frequency.
			rtcregs->rcnr = 0;
			rtcregs->rtar = 0xffffffff;
			rtcregs->rtsr |= RTSR_ale;
			rtcregs->rtsr |= RTSR_hze;
		}
		if (rtcregs->rcnr > savedtime)
			return rtcregs->rcnr - savedtime;
	} else
		savedtime = rtcregs->rcnr;
	clockinited = on;
	return 0L;
}
*/
void
clockinit(void)
{
	ulong x;
	ulong id;

	/* map the clock registers */
//	timerregs = mapspecial(OSTIMERREGS, sizeof(OSTimerRegs));
	intrenable(IRQ, IRQTC3OI, clockintr, nil, "clock");

	/* figure out processor frequency *
	x = powerregs->ppcr & 0x1f;
	conf.hz = ClockFreq*(x*4+16);
	conf.mhz = (conf.hz+499999)/1000000;
	*/
	conf.hz = 200*1024*1024;
	conf.mhz = 200*1024;

	/* get processor type */
	id = getcpuid();

	print("%lud MHZ ARM, ver %lux/part %lux/step %lud\n", conf.mhz,
		(id>>16)&0xff, (id>>4)&0xfff, id&0xf);

	/* post interrupt 1/HZ secs from now */
	when = timerregs->value + Minfreq;
	timerregs->load = when;

	/* enable RTC interrupts and alarms */
	intrenable(IRQ, IRQ64HZ, rtcintr, nil, "rtc");
	/* might want to use a different timer someday ... 64HZ is a lot. */

	timersinit();

	iprint("WE DID NOT ENABLE THE OS TIMER \n");
	clockinited = 1;
}

/*  turn 32 bit counter into a 64 bit one.  since todfix calls
 *  us at least once a second and we overflow once every 1165
 *  seconds, we won't miss an overflow.
 */
uvlong
fastticks(uvlong *hz)
{
	static uvlong high;
	static ulong last;
	ulong x;

	if(hz != nil)
		*hz = ClockFreq;
	x = timerregs->value;
	if(x < last)
		high += 1LL<<32;
	last = x;
	return high+x;
}

void
timerset(uvlong v)
{
	ulong next, tics;	/* Must be unsigned! */
	static int count;

	next = v;

	/* post next interrupt: calculate # of tics from now */
	tics = next - timerregs->value - Maxfreq;
	if (tics > Minfreq){
		timeradjust++;
		next = timerregs->value + Maxfreq;
	}
	timerregs->load = next;
}

static void
clockintr(Ureg *ureg, void*)
{
	/* reset previous interrupt */
	/* I am assuming we disable it here. this is sa100 code. 
	timerregs->ossr |= 1<<0;
	*/
	/* oh well we'll take our chances */
	when += Minfreq;
	timerregs->load = when;	/* insurance */

	timerintr(ureg, when);
}

/* ah shit. It wants a programmable RTC. Later */
void
rtcalarm(ulong secs)
{
	vlong t;

	if (t == 0){
		iprint("RTC alarm cancelled\n");
	} else {
		t = todget(nil);
		t = t / 1000000000ULL; // nsec to secs
		if (secs < t)
			return;
		secs -= t;
		iprint("RTC alarm set to %uld seconds from now\n", secs);
	}
}

static void
rtcintr(Ureg*, void*)
{
	unsigned long *teoi = (unsigned long *) 0x80930018;
	/* reset interrupt */
	/* need to write to TEOI */
	*teoi = 1;
	post(0xdd);
	iprint("RTC alarm:\n");
}

void
delay(int ms)
{
	ulong start;
	int i;

	if(clockinited){
		while(ms-- > 0){
			start = timerregs->value;
			while(timerregs->value-start < ClockFreq/1000)
				;
		}
	} else {
		while(ms-- > 0){
			for(i = 0; i < 1000; i++)
				;
		}
	}
}

void
microdelay(int µs)
{
	ulong start;
	int i;

	µs++;
	if(clockinited){
		start = timerregs->value;
		while(timerregs->value - start < 1UL+(µs*ClockFreq)/1000000UL)
			;
	} else {
		while(µs-- > 0){
			for(i = 0; i < 10; i++)
				;
		}
	}
}

/*  
 *  performance measurement ticks.  must be low overhead.
 *  doesn't have to count over a second.
 */
ulong
perfticks(void)
{
	return timerregs->value;
}
