#include	"u.h"
#include	"../port/lib.h"
#include	"mem.h"
#include	"dat.h"
#include	"fns.h"
#include	"io.h"
#include	"ureg.h"
#include	"../port/error.h"

/* we use TIMER1 for rtc, and timer 3 (32-bits) for OS timer */
/* the clock code seems to like a count-up forever timer, so 
 * we use clock timer 4 to track the time, and timer 3 to gen
 * the interrupt. We're having trouble with this, though, as 
 * the intervals we're asked to do are WAY too small -- 
 * order of 100 microseconds -- way too fast for this thing.
 * for now we're cheating and never letting it set a time
 * any faster than minfreq. 
 */
enum {
	Never	=	0xffffffff,
};

/* the RTC wil interrupt at 64 HZ. 
 * the OS timer is the 32-bit timer 3, running at 508 KHZ.
 */

OSTimerRegs *ostimer = (OSTimerRegs*)OSTIMERREGS;
OSTimerRegs *rtctimer = (OSTimerRegs *) RTCTIMERREGS;
struct timer4regs *timer4 = (struct timer4regs *) TIMER4REGS;
static int clockinited;

static void	clockintr(Ureg*, void*);
static void	rtcintr(Ureg*, void*);
static uvlong	when;	/* scheduled time of next interrupt */

long	timeradjust;

enum
{
	Minfreq = ClockFreq/HZ,		/* At least one interrupt per HZ (50 ms) */
	Maxfreq = ClockFreq/10000,	/* At most one interrupt every 100 µs */
};

/*
ulong
clockpower(int on)
{
	static ulong savedtime;

	if (on){
		timerregs->ossr |= 1<<0;
		timerregs->oier = 1<<0;
		timerregs->osmr[0] = timerregs->oscr + Minfreq;
		if (rtcregs->rttr == 0){
			rtcregs->rttr = 0x8000; // nominal frequency.
			rtcregs->rcnr = 0;
			rtcregs->rtar = 0xffffffff;
			rtcregs->rtsr |= RTSR_ale;
			rtcregs->rtsr |= RTSR_hze;
		}
		if (rtcregs->rcnr > savedtime)
			return rtcregs->rcnr - savedtime;
	} else
		savedtime = rtcregs->rcnr;
	clockinited = on;
	return 0L;
}
*/
void
clockinit(void)
{
	ulong x;
	ulong id;

	/* map the clock registers */


	/* figure out processor frequency *
	x = powerregs->ppcr & 0x1f;
	conf.hz = ClockFreq*(x*4+16);
	conf.mhz = (conf.hz+499999)/1000000;
	*/
	conf.hz = 200*1024*1024;
	conf.mhz = 200*1024;

	/* get processor type */
	id = getcpuid();

	print("%lud MHZ ARM, ver %lux/part %lux/step %lud\n", conf.mhz,
		(id>>16)&0xff, (id>>4)&0xfff, id&0xf);

	/* post interrupt 1/HZ secs from now */
	ostimer->control = 0;
	when = ostimer->value + Minfreq;
	iprint("setting load to 0x%lux\n", (ulong) when);
	ostimer->load = (ulong) when;
	intrenable(IRQ, IRQTC3OI, clockintr, nil, "clock");
	/* enable the 983Khz up counter */
	timer4->enable_valuehigh = TIMER4ENABLE;
	iprint("Enabled timer4, value low 0x%lux, high 0x%lux\n", 
			timer4->valuelow, timer4->enable_valuehigh);
	/* you can't program it to just stop when it hits zero. Only choice is 
	 * reload or wrap to -1. We elect for reload.
	 */
	ostimer->control = 0xc8; 	/* 508 Khz, periodic */

	/* enable RTC interrupts and alarms */
	intrenable(IRQ, IRQTC1OI, rtcintr, nil, "rtc");
	/* might want to use a different timer someday ... 64HZ is a lot. */
	rtctimer->control = 0;
	rtctimer->load = 10000; /* microseconds*2 = 20 milliseconds = 50 hz we hope */
	rtctimer->control = 0xc8; /* 508 Khz, periodic mode */

	timersinit();

	iprint("WE DID NOT ENABLE THE OS TIMER \n");
	clockinited = 1;
}

/*  turn 32 bit counter into a 64 bit one.  since todfix calls
 *  us at least once a second and we overflow once every 1165
 *  seconds, we won't miss an overflow.
 */
uvlong
fastticks(uvlong *hz)
{
	static uvlong high = 0;
	uvlong t4;
	static uvlong last = 0;
	ulong x;

	if(hz != nil)
		*hz = 983*1024;
	x = timer4->valuelow;
	t4 = timer4->enable_valuehigh & 0xff;
	t4 <<= 32;
	t4 |= x;
	/* this is bullshit. This timer only wraps 
	 * in 1,000,000 seconds. Any wrap here is 
	 * a hardware bug. We're seeing wraps in 
	 * practice. This hardware blows. 
	 * check my math: 
	 * 40 bits of timer, 20 bits of timer per second
	 * leaves 20 bits of seconds. 
	 */
	if(t4 < last){
//		high += 1LL<<40;
		iprint("TIMER WRAP! now 0x%llux was 0x%llux high 0x%llux\n", 
			t4, last, high);
	}
	last = t4;
	return high+t4;
}

void
timerset(uvlong v)
{
	uvlong next, tics;	/* Must be unsigned! */
	static int count;

	next = v;

	/* post next interrupt: calculate # of tics from now */
	tics = next - fastticks(0) - Maxfreq;
	if (tics > Minfreq){
		timeradjust++;
		tics = Maxfreq;
//		next = ostimer->value + Maxfreq;
	}
//	iprint("setting ostimer->load to 0x%llux\n", tics);
	ostimer->control = 0;

	/* note that timer 3 runs at half the rate of timer 2 */
//	ostimer->load = next;
	/* anything less than <<2 and this thing will hang. 
	 * no idea why. but procs never run!
	 */
	tics = Minfreq;
//	ostimer->load = next<<2;
	ostimer->load = (ulong)tics;
	ostimer->control = 0xc8;

}

static void
clockintr(Ureg *ureg, void*)
{
	/* reset previous interrupt */
	/* oh well we'll take our chances */
	/*kick teoi */
	ostimer->clear = 0;
//	poststr("O");
	when += Minfreq;
	ostimer->load = Minfreq; //(ulong) when;	/* insurance */

	timerintr(ureg, when);
//	iprint("DONE timerintr\n");
}

void
rtcalarm(ulong secs)
{
	vlong t;

	if (t == 0){
		iprint("RTC alarm cancelled\n");
	} else {
		t = todget(nil);
		t = t / 1000000000ULL; // nsec to secs
		if (secs < t)
			return;
		secs -= t;
		iprint("RTC alarm set to %uld seconds from now\n", secs);
	}
}

static void
rtcintr(Ureg*, void*)
{
	extern void sa930x_uartintr(Ureg*,void*);
	extern Uart sa930xuart[1];
	/* reset interrupt */
	/* need to write to TEOI */
	//iprint("rtcintr 0 to &%p\n", &rtctimer->clear);
	rtctimer->clear = 0;
	post(0xdd);
	/* work around IDIOTIC design of the uart interrupt ... */
	sa930x_uartintr(nil,&sa930xuart[0]);
//	iprint("RTC alarm:\n");
}

void
delay(int ms)
{
	ulong start;
	int i;

	if(clockinited){
		while(ms-- > 0){
			start = ostimer->value;
			while(ostimer->value-start < ClockFreq/1000)
				;
		}
	} else {
		while(ms-- > 0){
			for(i = 0; i < 1000; i++)
				;
		}
	}
}

void
microdelay(int µs)
{
	ulong start;
	int i;

	µs++;
	if(clockinited){
		start = ostimer->value;
		while(ostimer->value - start < 1UL+(µs*ClockFreq)/1000000UL)
			;
	} else {
		while(µs-- > 0){
			for(i = 0; i < 10; i++)
				;
		}
	}
}

/*  
 *  performance measurement ticks.  must be low overhead.
 *  doesn't have to count over a second.
 */
ulong
perfticks(void)
{
	return ostimer->value;
}
