

/*
 * GoMotion Veep Ethernet Controller.
 * This is just an example of the layout and structures
 * involved in getting a driver off the ground.
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "io.h"
#include "../port/error.h"
#include "../port/netif.h"

#include "etherif.h"
//#include "ethermii.h"
typedef struct D D;
typedef struct D {
	unsigned long e0, e1;
} D;

typedef struct S S;
typedef struct S {
	unsigned long e0;
} S;

typedef D receiveDescriptor;
typedef D receiveStatus;
typedef D transmitDescriptor;
typedef S transmitStatus;


/*---------------------------------------------------------------
 *  Data types used in this driver
 *-------------------------------------------------------------*/
typedef  unsigned char   U8;
typedef  unsigned short  U16;
typedef  unsigned long   U32;
typedef  unsigned int    UINT;
#ifdef DOSTRUCT
struct etherregs {
	ulong MAC_RXCTL; /*0x00  2-RW Rx  Control */
	ulong MAC_TXCTL; /*0x04  1-RW Tx Control */
	ulong MAC_TESTCTL; /*0x08  1-RW Test Control */
	ulong;
	ulong MAC_MIICMD; /*0x10  2-RW MII(Media Independent Intf) Command */
	ulong MAC_MIIDATA; /*0x14  2-RW MII Data */
	ulong MAC_MIISTS; /*0x18  1-RO MII Status */
	ulong;
	ulong MAC_SELFCTL; /*0x20  1-RW Self Control for LED interface */
	ulong MAC_INTEN; /*0x24  4-RW Intrrpt Enable */
	ulong MAC_INTSTSP; /*0x28  4-RW Intrrpt Status Preserve */
	ulong MAC_INTSTSC; /*0x2C  4-RO Intrrpt Status Clear */

	ulong MAC_DIAGAD; /*0x38  4-RW Diag Addr; /*debug only) */
	ulong MAC_DIAGDATA; /*0x3C  4-RW Diag Data; /*debug only) */

	ulong MAC_GT; /*0x40  4-RW General Timer */
	ulong MAC_FCT; /*0x44  4-RO Flow Control Timer */
	ulong MAC_FCF; /*0x48  4-RW Flow Control Format */
	ulong MAC_AFP; /*0x4C  1-RW Addr Filter Pointer */
	ulong MAC_HASHTB; /*0x50  8-RW Logical Addr Filter; /*Hash Table) */
	ulong MAC_INDAD; /*0x50  6-RW Individual Addr, IA */
	ulong MAC_INDAD_UPPER; /*0x54  6-RW Individual Addr, IA */

	ulong MAC_FER; /*0x60  4-RW Cardbus Functn Event Reg */
	ulong MAC_FERMASK; /*0x64  4-RW Cardbus Functn Event Mask Reg */
	ulong MAC_FPSR; /*0x68  4-RO Cardbus Functn Present Status Reg */
	ulong MAC_FFER; /*0x6C  4-RW Cardbus Functn Force Event Reg */
	ulong MAC_TXCOLLCNT; /*0x70  2-RW Tx Collision Count */
	ulong MAC_RXMISSCNT; /*0x74  2-RW Rx Miss Count */
	ulong MAC_RXRUNTCNT; /*0x78  2-RW Rx Runt Count */

	ulong MAC_BMCTL; /*0x80  1-RW Bus Master Control */
	ulong MAC_BMSTS; /*0x84  1-RO Bus Master Status */
	ulong MAC_RXBCA; /*0x88  4-RO Rx buffer current address */
	ulong MAC_TXBCA; /*0x8C  4-RO Tx buffer current address */
	ulong MAC_RXDBA; /*0x90  4-RW Rx Descrptr Queue Base Addr */
	ulong MAC_RXDBL; /*UINT16*)(0x94  2-RW Rx Descrptr Queue Base Length */
	ulong MAC_RXDCA; /*0x98  4-RW Rx Descrptr Current Addr */
	ulong MAC_RXDEQ; /*0x9C  2-RW Rx Descrptr Enqueue */ 
	ulong MAC_RXSBA; /*0xA0  4-RW Rx Status Queue Base Addr */
	ulong MAC_RXSBL; /*UINT16*)(0xA4  2-RW Rx Status Queue Base Length */

	ulong MAC_RXSCA; /*0xA8  4-RW Rx Status Current Addr */
	ulong MAC_RXSEQ; /*0xAC  2-RW Rx Status Enqueue */
	ulong MAC_TXDBA; /*0xB0  4-RW Tx Descrptr Queue Base Addr */
	ulong MAC_TXDBL; /*0xB4  2-RW Tx Descrptr Queue Base Length */
	ulong MAC_TXDCL; /*0xB6  2-RW Tx Descrptr Queue Current Length */
	ulong MAC_TXDCA; /*0xB8  4-RW Tx Descrptr Current Addr */
	ulong MAC_TXDEQ; /*0xBC  2-RW Tx Descrptr Enqueue */

	ulong MAC_TXSBA; /*0xC0  4-RW Tx status Queue Base Addr */
	ulong MAC_TXSBL; /*0xC4  2-RW Tx Status Queue Base Length */
	ulong MAC_TXSCL; /*0xC6  2-RW Tx Status Queue Current Length */
	ulong MAC_TXSCA; /*0xC8  4-RW Tx Status Current Addr */
	ulong MAC_TXSEQ; /*0xCC  4-RW Tx Status Current Addr */
	ulong MAC_RXBTH; /*0xD0  4-RW Rx Buffer Thrshold */
	ulong MAC_TXBTH; /*0xD4  4-RW Tx Buffer Thrshold */
	ulong MAC_RXSTH; /*0xD8  4-RW Rx Status Thrshold */
	ulong MAC_TXSTH; /*0xDC  4-RW Tx Status Thrshold */

	ulong MAC_RXDTH; /*0xE0  4-RW Rx Descrptr Thrshold */
	ulong MAC_TXDTH; /*0xE4  4-RW Tx Descrptr Thrshold */
	ulong MAC_MAXFL; /*0xE8  4-RW Maximum Frame Length */
	ulong MAC_RXHLEN; /*0xEC  2-RW Rx Header Length */
	ulong MAC_CFG_REG0; /*0x100  config registers 0-2 */
	ulong MAC_CFG_REG1; /*0x104    */
	ulong MAC_CFG_REG2; /*0x108    */
};

#endif
#define MAC_BASE				0x80010000
#define MAC_RXCTL               (MAC_BASE+0x00)  /* 2-RW Rx  Control */
#define MAC_TXCTL               (MAC_BASE+0x04)  /* 1-RW Tx Control */
#define MAC_TESTCTL             (MAC_BASE+0x08)  /* 1-RW Test Control */
#define MAC_MIICMD              (MAC_BASE+0x10)  /* 2-RW MII(Media Independent Intf) Command */
#define MAC_MIIDATA             (MAC_BASE+0x14)  /* 2-RW MII Data */
#define MAC_MIISTS              (MAC_BASE+0x18)  /* 1-RO MII Status */

#define MAC_SELFCTL             (MAC_BASE+0x20)  /* 1-RW Self Control for LED interface */
#define MAC_INTEN               (MAC_BASE+0x24)  /* 4-RW Intrrpt Enable */
#define MAC_INTSTSP             (MAC_BASE+0x28)  /* 4-RW Intrrpt Status Preserve */
#define MAC_INTSTSC             (MAC_BASE+0x2C)  /* 4-RO Intrrpt Status Clear */

#define MAC_DIAGAD              (MAC_BASE+0x38)  /* 4-RW Diag Addr (debug only) */
#define MAC_DIAGDATA            (MAC_BASE+0x3C)  /* 4-RW Diag Data (debug only) */

#define MAC_GT                  (MAC_BASE+0x40)  /* 4-RW General Timer */
#define MAC_FCT                 (MAC_BASE+0x44)  /* 4-RO Flow Control Timer */
#define MAC_FCF                 (MAC_BASE+0x48)  /* 4-RW Flow Control Format */
#define MAC_AFP                 (MAC_BASE+0x4C)  /* 1-RW Addr Filter Pointer */
#define MAC_HASHTB              (MAC_BASE+0x50)  /* 8-RW Logical Addr Filter (Hash Table) */
#define MAC_INDAD               (MAC_BASE+0x50)  /* 6-RW Individual Addr, IA */
#define MAC_INDAD_UPPER         (MAC_BASE+0x54)  /* 6-RW Individual Addr, IA */

#define MAC_FER                 (MAC_BASE+0x60)  /* 4-RW Cardbus Functn Event Reg */
#define MAC_FERMASK             (MAC_BASE+0x64)  /* 4-RW Cardbus Functn Event Mask Reg */
#define MAC_FPSR                (MAC_BASE+0x68)  /* 4-RO Cardbus Functn Present Status Reg */
#define MAC_FFER                (MAC_BASE+0x6C)  /* 4-RW Cardbus Functn Force Event Reg */
#define MAC_TXCOLLCNT           (MAC_BASE+0x70)  /* 2-RW Tx Collision Count */
#define MAC_RXMISSCNT           (MAC_BASE+0x74)  /* 2-RW Rx Miss Count */
#define MAC_RXRUNTCNT           (MAC_BASE+0x78)  /* 2-RW Rx Runt Count */

#define MAC_BMCTL               (MAC_BASE+0x80)  /* 1-RW Bus Master Control */
#define MAC_BMSTS               (MAC_BASE+0x84)  /* 1-RO Bus Master Status */
#define MAC_RXBCA               (MAC_BASE+0x88)  /* 4-RO Rx buffer current address */
#define MAC_TXBCA               (MAC_BASE+0x8C)  /* 4-RO Tx buffer current address */
#define MAC_RXDBA               (MAC_BASE+0x90)  /* 4-RW Rx Descrptr Queue Base Addr */
#define MAC_RXDBL               (UINT16*)(MAC_BASE+0x94)  /* 2-RW Rx Descrptr Queue Base Length */
#define MAC_RXDCA               (MAC_BASE+0x98)  /* 4-RW Rx Descrptr Current Addr */
#define MAC_RXDEQ               (MAC_BASE+0x9C)  /* 2-RW Rx Descrptr Enqueue */ 
#define MAC_RXSBA               (MAC_BASE+0xA0)  /* 4-RW Rx Status Queue Base Addr */
#define MAC_RXSBL               (UINT16*)(MAC_BASE+0xA4)  /* 2-RW Rx Status Queue Base Length */

#define MAC_RXSCA               (MAC_BASE+0xA8)  /* 4-RW Rx Status Current Addr */
#define MAC_RXSEQ               (MAC_BASE+0xAC)  /* 2-RW Rx Status Enqueue */
#define MAC_TXDBA               (MAC_BASE+0xB0)  /* 4-RW Tx Descrptr Queue Base Addr */
#define MAC_TXDBL               (MAC_BASE+0xB4)  /* 2-RW Tx Descrptr Queue Base Length */
#define MAC_TXDCL               (MAC_BASE+0xB6)  /* 2-RW Tx Descrptr Queue Current Length */
#define MAC_TXDCA               (MAC_BASE+0xB8)  /* 4-RW Tx Descrptr Current Addr */
#define MAC_TXDEQ               (MAC_BASE+0xBC)  /* 2-RW Tx Descrptr Enqueue */

#define MAC_TXSBA               (MAC_BASE+0xC0)  /* 4-RW Tx status Queue Base Addr */
#define MAC_TXSBL               (MAC_BASE+0xC4)  /* 2-RW Tx Status Queue Base Length */
#define MAC_TXSCL               (MAC_BASE+0xC6)  /* 2-RW Tx Status Queue Current Length */
#define MAC_TXSCA               (MAC_BASE+0xC8)  /* 4-RW Tx Status Current Addr */
#define MAC_TXSEQ               (MAC_BASE+0xCC)  /* 4-RW Tx Status Current Addr */
#define MAC_RXBTH               (MAC_BASE+0xD0)  /* 4-RW Rx Buffer Thrshold */
#define MAC_TXBTH               (MAC_BASE+0xD4)  /* 4-RW Tx Buffer Thrshold */
#define MAC_RXSTH               (MAC_BASE+0xD8)  /* 4-RW Rx Status Thrshold */
#define MAC_TXSTH               (MAC_BASE+0xDC)  /* 4-RW Tx Status Thrshold */

#define MAC_RXDTH               (MAC_BASE+0xE0)  /* 4-RW Rx Descrptr Thrshold */
#define MAC_TXDTH               (MAC_BASE+0xE4)  /* 4-RW Tx Descrptr Thrshold */
#define MAC_MAXFL               (MAC_BASE+0xE8)  /* 4-RW Maximum Frame Length */
#define MAC_RXHLEN              (MAC_BASE+0xEC)  /* 2-RW Rx Header Length */
#define MAC_CFG_REG0            (MAC_BASE+0x100) /* config registers 0-2 */
#define MAC_CFG_REG1            (MAC_BASE+0x104) /*   */
#define MAC_CFG_REG2            (MAC_BASE+0x108) /*   */


/*---------------------------------------------------------------
 * Definition of the registers.
 * For details, refer to the datasheet  .
 * 
 * Basically, most registers are 32 bits width register.
 * But some are 16 bits and some are 6 or 8 bytes long.
 *-------------------------------------------------------------*/

#define  REG_RxCTL  0x0000  /*offset to Receiver Control Reg*/
#define  RxCTL_PauseA  (1<<20)
#define  RxCTL_RxFCE1  (1<<19)
#define  RxCTL_RxFCE0  (1<<18)
#define  RxCTL_BCRC    (1<<17)
#define  RxCTL_SRxON   (1<<16)
#define  RxCTL_RCRCA   (1<<13)
#define  RxCTL_RA      (1<<12)
#define  RxCTL_PA      (1<<11)
#define  RxCTL_BA      (1<<10)
#define  RxCTL_MA      (1<<9)
#define  RxCTL_IAHA    (1<<8)
#define  RxCTL_IA3     (1<<3)
#define  RxCTL_IA2     (1<<2)
#define  RxCTL_IA1     (1<<1)
#define  RxCTL_IA0     (1<<0)

#define  REG_TxCTL  0x0004  /*offset to Transmit Control Reg*/
#define  TxCTL_DefDis  (1<<7)
#define  TxCTL_MBE     (1<<6)
#define  TxCTL_ICRC    (1<<5)
#define  TxCTL_TxPD    (1<<5)
#define  TxCTL_OColl   (1<<3)
#define  TxCTL_SP      (1<<2)
#define  TxCTL_PB      (1<<1)
#define  TxCTL_STxON   (1<<0)

#define  REG_TestCTL   0x0008  /*Test Control Reg, R/W*/
#define  TestCTL_MACF  (1<<7)
#define  TestCTL_MFDX  (1<<6)
#define  TestCTL_DB    (1<<5)
#define  TestCTL_MIIF  (1<<4)

#define  REG_MIICmd  0x0010  /*offset to MII Command Reg, R/W*/
#define  MIICmd_OP     (0x03<<14)
#define  MIICmd_OP_RD  (2<<14)
#define  MIICmd_OP_WR  (1<<14)
#define  MIICmd_PHYAD  (0x1f<<5)
#define  MIICmd_REGAD  (0x1f<<0)

#define  REG_MIIData  0x0014  /*offset to MII Data Reg, R/W*/
#define  MIIData_MIIData  (0xffff<<0)

#define  REG_MIISts  0x0018  /*offset to MII Status Reg, R*/
#define  MIISts_Busy  (1<<0)

#define  REG_SelfCTL  0x0020  /*offset to Self Control Reg*/
#define  SelfCTL_RWP    (1<<7)    /*Remote Wake Pin*/
#define  SelfCTL_GPO0   (1<<5)
#define  SelfCTL_PUWE   (1<<4)
#define  SelfCTL_PDWE   (1<<3)
#define  SelfCTL_MIIL   (1<<2)
#define  SelfCTL_RESET  (1<<0)

#define  REG_IntEn   0x0024  /*Interrupt Enable Reg, R/W*/
#define  IntEn_RWIE    (1<<30)
#define  IntEn_RxMIE   (1<<29)
#define  IntEn_RxBIE   (1<<28)
#define  IntEn_RxSQIE  (1<<27)
#define  IntEn_TxLEIE  (1<<26)
#define  IntEn_ECIE    (1<<25)
#define  IntEn_TxUHIE  (1<<24)
#define  IntEn_MOIE    (1<<18)
#define  IntEn_TxCOIE  (1<<17)
#define  IntEn_RxROIE  (1<<16)
#define  IntEn_MIIIE   (1<<12)
#define  IntEn_PHYSIE  (1<<11)
#define  IntEn_TIE     (1<<10)
#define  IntEn_SWIE    (1<<8)
#define  IntEn_TxSQIE   (1<<3)
#define  IntEn_RxEOFIE  (1<<2)
#define  IntEn_RxEOBIE  (1<<1)
#define  IntEn_RxHDRIE  (1<<0)

#define  REG_IntStsP  0x0028  /*offset to Interrupt Status Preserve Reg, R/W*/
#define  REG_IntStsC  0x002c  /*offset to Interrupt Status Clear Reg, R*/
#define  IntSts_RWI    (1<<30)
#define  IntSts_RxMI   (1<<29)
#define  IntSts_RxBI   (1<<28)
#define  IntSts_RxSQI  (1<<27)
#define  IntSts_TxLEI  (1<<26)
#define  IntSts_ECI    (1<<25)
#define  IntSts_TxUHI  (1<<24)
#define  IntSts_MOI    (1<<18)
#define  IntSts_TxCOI  (1<<17)
#define  IntSts_RxROI  (1<<16)
#define  IntSts_MIII   (1<<12)
#define  IntSts_PHYSI  (1<<11)
#define  IntSts_TI     (1<<10)
#define  IntSts_AHBE   (1<<9)
#define  IntSts_SWI    (1<<8)
#define  IntSts_OTHER  (1<<4)
#define  IntSts_TxSQ   (1<<3)
#define  IntSts_RxSQ   (1<<2)

#define  REG_GT  0x0040  /*offset to General Timer Reg*/
#define  GT_GTC  (0xffff<<16)
#define  GT_GTP  (0xffff<<0)

#define  REG_FCT  0x0044  /*offset to Flow Control Timer Reg*/
#define  FCT_FCT  (0x00ffffff<<0)

#define  REG_FCF  0x0048  /*offset to Flow Control Format Reg*/
#define  FCF_MACCT  (0xffff<<16)
#define  FCF_TPT    (0xffff<<0)

#define  REG_AFP  0x004c  /*offset to Address Filter Pointer Reg*/
#define  AFP_AFP  (0x07<<0)  /*Address Filter Pointer (bank control for REG_IndAD)*/
#define  AFP_AFP_IA0   0       /*Primary Individual Address (MAC Addr)*/
#define  AFP_AFP_IA1   1       /*Individual Address 1*/
#define  AFP_AFP_IA2   2       /*Individual Address 2*/
#define  AFP_AFP_IA3   3       /*Individual Address 3*/
#define  AFP_AFP_DTxP  6       /*Destination Address of Tx Pause Frame*/
#define  AFP_AFP_HASH  7       /*Hash Table*/

#define  REG_IndAD      0x0050  /*offset to Individual Address Reg, n bytes, R/W*/

#define  REG_GIntSts    0x0060  /*offset to Global Interrupt Status Reg (writing 1 will clear)*/
#define  REG_GIntROS    0x0068  /*offset to Global Interrupt Status Read Only Reg*/
#define  GIntSts_INT  (1<<15)     /*Global Interrupt Request Status*/

#define  REG_GIntMsk    0x0064  /*offset to Global Interrupt Mask Reg*/
#define  GIntMsk_IntEn  (1<<15)   /*Global Interrupt Enable*/

#define  REG_GIntFrc    0x006c  /*offset to Global Interrupt Force Reg*/
#define  GIntFrc_INT  (1<<15)     /*Force to set GIntSts*/

#define  REG_TxCollCnt  0x0070  /*Transmit Collision Count Reg, R*/
#define  REG_RxMissCnt  0x0074  /*Receive Miss Count Reg, R*/
#define  REG_RxRntCnt   0x0078  /*Receive Runt Count Reg, R*/

#define  REG_BMCtl  0x0080  /*offset to Bus Master Control Reg, R/W*/
#define  BMCtl_MT     (1<<13)
#define  BMCtl_TT     (1<<12)
#define  BMCtl_UnH    (1<<11)
#define  BMCtl_TxChR  (1<<10)
#define  BMCtl_TxDis  (1<<9)
#define  BMCtl_TxEn   (1<<8)
#define  BMCtl_EH2    (1<<6)
#define  BMCtl_EH1    (1<<5)
#define  BMCtl_EEOB   (1<<4)
#define  BMCtl_RxChR  (1<<2)
#define  BMCtl_RxDis  (1<<1)
#define  BMCtl_RxEn   (1<<0)

#define  REG_BMSts  0x0084  /*offset to Bus Master Status Reg, R*/
#define  BMSts_TxAct  (1<<7)
#define  BMSts_TP     (1<<4)
#define  BMSts_RxAct  (1<<3)
#define  BMSts_QID    (0x07<<0)
#define  BMSts_QID_RxDt   (0<<0)
#define  BMSts_QID_TxDt   (1<<0)
#define  BMSts_QID_RxSts  (2<<0)
#define  BMSts_QID_TxSts  (3<<0)
#define  BMSts_QID_RxDesc (4<<0)
#define  BMSts_QID_TxDesc (5<<0)

#define  REG_RBCA   0x0088  /*offset to Receive Buffer Current Address Reg, R*/
#define  REG_TBCA   0x008c  /*offset to Transmit Buffer Current Address Reg, R*/

#define  REG_RxDBA  0x0090  /*offset to Receive Descriptor Queue Base Address Reg, R/W*/
#define  REG_RxDBL  0x0094  /*offset to Receive Descriptor Queue Base Length Reg, R/W, 16bits*/
#define  REG_RxDCL  0x0096  /*offset to Receive Descriptor Queue Current Length Reg, R/W, 16bits*/
#define  REG_RxDCA  0x0098  /*offset to Receive Descriptor Queue Current Address Reg, R/W*/

#define  REG_RxDEQ  0x009c  /*offset to Receive Descriptor Enqueue Reg, R/W*/
#define  RxDEQ_RDV  (0xffff<<16)  /*R 16bit; Receive Descriptor Value*/
#define  RxDEQ_RDI  (0xff<<0)     /*W 8bit; Receive Descriptor Increment*/

#define  REG_RxSBA  0x00a0  /*offset to Receive Status Queue Base Address Reg, R/W*/
#define  REG_RxSBL  0x00a4  /*offset to Receive Status Queue Base Length Reg, R/W, 16bits*/
#define  REG_RxSCL  0x00a6  /*offset to Receive Status Queue Current Length Reg, R/W, 16bits*/
#define  REG_RxSCA  0x00a8  /*offset to Receive Status Queue Current Address Reg, R/W*/

#define  REG_RxSEQ  0x00ac  /*offset to Receive Status Queue Current Address Reg, R/W*/
#define  RxSEQ_RSV  (0xffff<<16)
#define  RxSEQ_RSI  (0xff<<0)

#define  REG_TxDBA  0x00b0  /*offset to Transmit Descriptor Queue Base Address Reg, R/W*/
#define  REG_TxDBL  0x00b4  /*offset to Transmit Descriptor Queue Base Length Reg, R/W, 16bits*/
#define  REG_TxDCL  0x00b6  /*offset to Transmit Descriptor Queue Current Length Reg, R/W, 16bits*/
#define  REG_TxDCA  0x00b8  /*offset to Transmit Descriptor Queue Current Address Reg, R/W*/

#define  REG_TxDEQ  0x00bc  /*offset to Transmit Descriptor Queue Current Address Reg, R/W*/
#define  TxDEQ_TDV  (0xffff<<16)
#define  TxDEQ_TDI  (0xff<<0)

#define  REG_TxSBA  0x00c0  /*offset to Transmit Status Queue Base Address Reg, R/W*/
#define  REG_TxSBL  0x00c4  /*offset to Transmit Status Queue Base Length Reg, R/W, 16bits*/
#define  REG_TxSCL  0x00c6  /*offset to Transmit Status Queue Current Length Reg, R/W, 16bits*/
#define  REG_TxSCA  0x00c8  /*offset to Transmit Status Queue Current Address Reg, R/W*/

#define  REG_RxBTH  0x00d0  /*offset to Receive Buffer Threshold Reg, R/W*/
#define  RxBTH_RDHT  (0x03ff<<16)
#define  RxBTH_RDST  (0x03ff<<0)

#define  REG_TxBTH  0x00d4  /*offset to Transmit Buffer Threshold Reg, R/W*/
#define  TxBTH_TDHT  (0x03ff<<16)
#define  TxBTH_TDST  (0x03ff<<0)

#define  REG_RxSTH  0x00d8  /*offset to Receive Status Threshold Reg, R/W*/
#define  RxSTH_RSHT  (0x003f<<16)
#define  RxSTH_RSST  (0x003f<<0)

#define  REG_TxSTH  0x00dc  /*offset to Transmit Status Threshold Reg, R/W*/
#define  TxSTH_TSHT  (0x003f<<16)
#define  TxSTH_TSST  (0x003f<<0)

#define  REG_RxDTH  0x00e0  /*offset to Receive Descriptor Threshold Reg, R/W*/
#define  RxDTH_RDHT  (0x003f<<16)
#define  RxDTH_RDST  (0x003f<<0)

#define  REG_TxDTH  0x00e4  /*offset to Transmit Descriptor Threshold Reg, R/W*/
#define  TxDTH_TDHT  (0x003f<<16)
#define  TxDTH_TDST  (0x003f<<0)

#define  REG_MaxFL  0x00e8  /*offset to Max Frame Length Reg, R/W*/
#define  MaxFL_MFL  (0x07ff<<16)
#define  MaxFL_TST  (0x07ff<<0)

#define  REG_RxHL  0x00ec  /*offset to Receive Header Length Reg, R/W*/
#define  RxHL_RHL2  (0x07ff<<16)
#define  RxHL_RHL1  (0x03ff<<0)

#define  REG_MACCFG0  0x0100  /*offset to Test Reg #0, R/W*/
#define  MACCFG0_DbgSel  (1<<7)
#define  MACCFG0_LCKEN   (1<<6)
#define  MACCFG0_LRATE   (1<<5)
#define  MACCFG0_RXERR   (1<<4)
#define  MACCFG0_BIT33   (1<<2)
#define  MACCFG0_PMEEN   (1<<1)
#define  MACCFG0_PMEST   (1<<0)

#define  REG_MACCFG1  0x0104  /*offset to Test Reg #1, R/W*/
#define  REG_MACCFG2  0x0108  /*offset to Test Reg #2, R*/
#define  REG_MACCFG3  0x010c  /*offset to Test Reg #3, R*/

typedef struct bufferDescriptor {
	void *vaddr;
} bufferDescriptor;

typedef  struct ep9213Eth_info  {
    /****  static device informations  **********************************/
    struct {
        int                 id;            /*device instance ID (0 for 1st and so on)
                                             must be first element of this structure*/
        D   *pQueRxDesc;   /*pointer to Rx Descriptor Queue*/
        D       *pQueRxSts;    /*pointer to Rx Status Queue*/
        D  *pQueTxDesc;   /*pointer to Tx Descriptor Queue*/
        S      *pQueTxSts;    /*pointer to Tx Status Queue*/
        unsigned char       *pRxBuf;       /*base of Rx Buffer pool*/
        unsigned char       *pTxBuf;       /*base of Tx Buffer pool*/
        unsigned long       phyQueueBase;  /*physical address of device queues*/
        unsigned long       phyQueRxDesc,  /*physical address of Rx Descriptor Queue*/
                            phyQueRxSts,   /*physical address of Rx Status Queue*/
                            phyQueTxDesc,  /*physical address of Tx Descriptor Queue*/
                            phyQueTxSts,   /*physical address of Tx Status Queue*/
                            phyRxBuf,      /*physical address of Rx Buffer pool*/
                            phyTxBuf;      /*physical address of Tx Buffer pool*/
		bufferDescriptor *pRxBufDesc,
						 *pTxBufDesc;
        int                 miiIdPhy;      /*MII Bus ID of Ethernet PHY*/
    }  s;
    /****  dynamic information, subject to clear when device open  ******/
    struct {
 //       struct net_device_stats  stats;  /*statistic data*/
        int  idxQueRxDesc,       /*next processing index of device queues*/
             idxQueRxSts,
             idxQueTxDescHead,
             idxQueTxDescTail,
             idxQueTxSts;
        int  txStopped;          /*flag for Tx condition*/
    }  d;
}  ep9213Eth_info;
/*---------------------------------------------------------------
 * Definition of Descriptor/Status Queue Entry
 *-------------------------------------------------------------*/

#ifdef BITFIELDSUCK
typedef  union receiveDescriptor  {  /*data structure of Receive Descriptor Queue Entry*/
    struct {        /*whole value*/
        U32  e0,      /*1st dword entry*/
             e1;      /*2nd dword entry*/
    }  w;
    struct {        /*bit field definitions*/
        U32  ba:32,   /*Buffer Address (keep in mind this is physical address)*/
             bl:16,   /*b15-0; Buffer Length*/
             bi:15,   /*b30-16; Buffer Index*/
             nsof:1;  /*b31; Not Start Of Frame*/
    }  f;
}  receiveDescriptor;


typedef  union receiveStatus  {  /*data structure of Receive Status Queue Entry*/
    struct {         /*whole word*/
        U32  e0,       /*1st dword entry*/
             e1;       /*2nd dword entry*/
    }  w;
    struct {         /*bit field*/
        U32  rsrv1:8,  /*b7-0: reserved*/
             hti:6,    /*b13-8: Hash Table Index*/
             rsrv2:1,  /*b14: reserved*/
             crci:1,   /*b15: CRC Included*/
             crce:1,   /*b16: CRC Error*/
             edata:1,  /*b17: Extra Data*/
             runt:1,   /*b18: Runt Frame*/
             fe:1,     /*b19: Framing Error*/
             oe:1,     /*b20: Overrun Error*/
             rxerr:1,  /*b21: Rx Error*/
             am:2,     /*b23-22: Address Match*/
             rsrv3:4,  /*b27-24: reserved*/
             eob:1,    /*b28: End Of Buffer*/
             eof:1,    /*b29: End Of Frame*/
             rwe:1,    /*b30: Received Without Error*/
             rfp:1,    /*b31: Receive Frame Processed*/
             fl:16,    /*b15-0: frame length*/
             bi:15,    /*b30-16: Buffer Index*/
             rfp2:1;   /*b31: Receive Frame Processed at 2nd word*/
    }  f;
}  receiveStatus;


typedef  union transmitDescriptor  {  /*data structure of Transmit Descriptor Queue Entry*/
    struct {         /*whole value*/
        U32  e0,       /*1st dword entry*/
             e1;       /*2nd dword entry*/
    }  w;
    struct {         /*bit field*/
        U32  ba:32,    /*b31-0: Buffer Address (keep in mind this is physical address)*/
             bl:12,    /*b11-0: Buffer Length*/
             rsrv1:3,  /*b14-12: reserved*/
             af:1,     /*b15: Abort Frame*/
             bi:15,    /*b30-16: Buffer Index*/
             eof:1;    /*b31: End Of Frame*/

    }  f;
}  transmitDescriptor;


typedef  union transmitStatus  {  /*data structure of Transmit Status Queue Entry*/
    struct {         /*whole word*/
        U32  e0;       /*1st dword entry*/
    }  w;
    struct {         /*bit field*/
        U32  bi:15,    /*b14-0: Buffer Index*/
             rsrv3:1,  /*b15: reserved*/
             ncoll:5,  /*b20-16: Number of Collisions*/
             rsrv2:3,  /*b23-21: reserved*/
             ecoll:1,  /*b24: Excess Collisions*/
             txu:1,    /*b25: Tx Underrun*/
             ow:1,     /*b26: Out of Window*/
             rsrv1:1,  /*b27: reserved*/
             lcrs:1,   /*b28: Loss of CRS*/
             fa:1,     /*b29: Frame Abort*/
             txwe:1,   /*b30: Transmitted Without Error*/
             txfp:1;   /*b31: Transmit Frame Processed*/
    }  f;
}  transmitStatus;

#endif


/*---------------------------------------------------------------
 *  Size of device registers occupied in memory/IO address map
 *-------------------------------------------------------------*/
#define  DEV_REG_SPACE  0x00010000

/*
#define U8 unsigned char
#define U16 unsigned short
#define U32 unsigned long
*/


/*----------------------------------------------------------------------------
 * DEBUGGING LEVELS
 *
 * 0 for normal operation
 * 1 for slightly more details
 * >2 for various levels of increasingly useless information
 *    2 for interrupt tracking, status flags
 *    3 for packet dumps, etc.
 *--------------------------------------------------------------------------*/
#undef DBG2 
#undef DBG1 
#undef DBG0 
#define DBG 

#ifdef DBG2
#define PRINTK3(x) iprint x
#else
#define PRINTK3(x)
#endif

#ifdef DBG1
#define PRINTK2(x) iprint x
#else
#define PRINTK2(x)
#endif

#ifdef DBG0
#define PRINTK1(x) iprint x
#else
#define PRINTK1(x)
#endif

#ifdef DBG
#define PRINTK(x) iprint x
#else
#define PRINTK(x)
#endif

#define  _PRTK_ENTRY      PRINTK2   /*to trace function entries*/
#define  _PRTK_SWERR      PRINTK    /*logical S/W error*/
#define  _PRTK_SYSFAIL    PRINTK    /*system service failure*/
#define  _PRTK_HWFAIL     PRINTK    /*H/W operation failure message*/
#define  _PRTK_WARN       PRINTK1   /*warning information*/
#define  _PRTK_INFO       PRINTK2   /*general information*/
#define  _PRTK_ENTRY_ISR  PRINTK3   /*to trace function entries belong to ISR*/
#define  _PRTK_WARN_ISR   PRINTK1   /*warning informations from ISR*/
#define  _PRTK_INFO_ISR   PRINTK3   /*general informations from ISR*/
#define  _PRTK_           PRINTK    /*for temporary print out*/

#define  _PRTK_DUMP       PRINTK1   /*to dump large amount of debug info*/

/* this kind of macro is stupid, it will all go away soon, 
 * i am killing it bit-by-bit  -- RGM
 */
/*---------------------------------------------------------------
 * A definition of register access macros
 *-------------------------------------------------------------*/
#define  _RegRd(type,ofs)     (*(volatile type*)(ofs))
#define  _RegWr(type,ofs,dt)  *(volatile type*)(ofs)=((type)(dt))
#define  RegRd8(ofs)   _RegRd(U8,(char*)pD->base_addr+(ofs))
#define  RegRd16(ofs)  _RegRd(U16,(char*)pD->base_addr+(ofs))
#define  aRegRd32(ofs)  _RegRd(U32,(char*)pD->base_addr+(ofs))
#define  RegWr8(ofs,dt)   _RegWr(U8,(char*)pD->base_addr+(ofs),(dt))
#define  RegWr16(ofs,dt)  _RegWr(U16,(char*)pD->base_addr+(ofs),(dt))
#define  aRegWr32(ofs,dt)  _RegWr(U32,(char*)pD->base_addr+(ofs),(dt))

#define  RegRd32(ofs)  _RegRd32((char*)pD->base_addr+(ofs))
#define  RegWr32(ofs,dt)  _RegWr32((char*)pD->base_addr+(ofs),(dt))

static ulong watchIO = 0;
void
_RegWr32(volatile char *address, ulong value) {
	if (watchIO)
		iprint("W32 %p=0x%lux\n", address, value);
	*(volatile ulong *)address = value;
	/* we'd like to check to make sure it got it but:
	 * reads may not be side-effect free
	 * not all bits may come back. 
	 * so no check -- RGM
	 */
}

ulong
_RegRd32(volatile char *address) {
	ulong value = * (volatile ulong *) address;
	if (watchIO)
		iprint("R32 %p=0x%lux\n", address, value);
	return value;
}


/*----------------------------------------------------------------------------
 * Some definitions belong to the operation of this driver.
 * You should understand how it affect to driver before any modification.
 *--------------------------------------------------------------------------*/

/****  Interrupt Sources in Use  *******************************************/
/*#define  Default_IntSrc  (IntEn_RxMIE|IntEn_RxSQIE|IntEn_TxLEIE|IntEn_TIE|IntEn_TxSQIE|IntEn_RxEOFIE|IntEn_RxEOBIE|IntEn_RxHDRIE)
*/
#define  Default_IntSrc  (IntEn_TxSQIE|IntEn_RxEOFIE|IntEn_RxEOBIE|IntEn_RxHDRIE)

/****  Length of Device Queue in number of entries
       (must be less than or equal to 255)  ********************************/
#define  LEN_QueRxDesc  64             /*length of Rx Descriptor Queue (4 or bigger) Must be power of 2.*/
#define  LEN_QueRxSts   LEN_QueRxDesc  /*length of Rx Status Queue*/
#define  LEN_QueTxDesc  8              /*length of Tx Descriptor Queue (4 or bigger) Must be power of 2.*/
#define  LEN_QueTxSts   LEN_QueTxDesc  /*length of Tx Status Queue*/

/****  Tx Queue fill-up level control  *************************************/
#define  LVL_TxStop    LEN_QueTxDesc - 2    /*level to ask the stack to stop Tx*/
#define  LVL_TxResume  2   /*level to ask the stack to resume Tx*/

/****  Rx Buffer length in bytes  ******************************************/
#define  LEN_RxBuf  (1518+2+16)  /*length of Rx buffer, must be 4-byte aligned*/
#define  LEN_TxBuf   LEN_RxBuf


/*----------------------------------------------------------------------------
 * MACRO for ease
 *--------------------------------------------------------------------------*/

#define  Align32(a)  (((unsigned int)(a)+3)&~0x03)  /*32bit address alignment*/
#define  IdxNext(idxCur,len)  (((idxCur)+1)%(len))  /*calc next array index number*/

typedef struct Ctlr Ctlr;
typedef struct Ctlr {
	Ether *edev;
	unsigned char *base_addr;
	D *rd, *td;
	char dev_addr[6];
	int	port;
	Pcidev*	pcidev;
	Ctlr*	next;
	int	active;
	uint	id;

	QLock	alock;			/* attach */
	Lock	ilock;			/* init */
	int	init;			/*  */

//	Mii*	mii;

	Lock	tlock;			/* transmit */
//	D*	td;			/* descriptor ring */
	Block**	tb;			/* transmit buffers */
	int	ntd;

	int	tdh;			/* head - producer index (host) */
	int	tdt;			/* tail - consumer index (NIC) */
	int	ntdfree;
	int	ntq;

	int	mtps;			/* Max. Transmit Packet Size */

	Lock	rlock;			/* receive */
//	D*	rd;			/* descriptor ring */
	Block**	rb;			/* receive buffers */
	int	nrd;

	int	rdh;			/* head - producer index (NIC) */
	int	rdt;			/* tail - consumer index (host) */
	int	nrdfree;

	int	tcr;			/* transmit configuration register */
	int	rcr;			/* receive configuration register */

	QLock	slock;			/* statistics */
	uint	txdu;
	uint	tcpf;
	uint	udpf;
	uint	ipf;
	uint	fovf;
	uint	ierrs;
	uint	rer;
	uint	rdu;
	uint	punlc;
	uint	fovw;
	ep9213Eth_info priv;
} Ctlr;

/* MALLOC_DMA */
/* Keep all the dma memory in the non-coherent space */
unsigned char *dmamem = (unsigned char *) 0xc0100000;
#ifdef NO
void *
MALLOC_DMA(int amount, unsigned long *pptr) {
	void *retval;
	retval = malloc(amount);
	if (retval && pptr)
		*pptr = PADDR(retval);
	return retval;
}
#endif
void *
MALLOC_DMA(int amount, unsigned long *pptr) {
	void *retval;
	iprint("malloc dma, for %d bytes\n", amount);
	retval = dmamem;
	dmamem += amount;
	if (dmamem > (uchar *) 0xc0200000)
		panic("out of dma mem\n");
	if (retval && pptr)
		*pptr = PADDR(retval);
	return retval;
}
static int  eth_isrTx(struct Ctlr *pD);
static void	ep93xxinterrupt(Ureg*, void* arg);
#define TS7XXX_FLASH_BASE 0x60000000
#define IO_ADDRESS(x) x
// TS-7200 may have 8 or 16 mb flash
#define  GET_MAC_ADDR(pD, pMac)	{ \
if (*(unsigned int *)(IO_ADDRESS(TS7XXX_FLASH_BASE) + 0xe20000) == 0x53555243) { \
	(pMac)[0] = *(unsigned char *)(IO_ADDRESS(TS7XXX_FLASH_BASE) + 0xfc030a); \
	(pMac)[1] = *(unsigned char *)(IO_ADDRESS(TS7XXX_FLASH_BASE) + 0xfc030b); \
	(pMac)[2] = *(unsigned char *)(IO_ADDRESS(TS7XXX_FLASH_BASE) + 0xfc030c); \
	(pMac)[3] = *(unsigned char *)(IO_ADDRESS(TS7XXX_FLASH_BASE) + 0xfc030d); \
	(pMac)[4] = *(unsigned char *)(IO_ADDRESS(TS7XXX_FLASH_BASE) + 0xfc030e); \
	(pMac)[5] = *(unsigned char *)(IO_ADDRESS(TS7XXX_FLASH_BASE) + 0xfc030f); \
} else { \
	(pMac)[0] = *(unsigned char *)(IO_ADDRESS(TS7XXX_FLASH_BASE) + 0x7c030a); \
	(pMac)[1] = *(unsigned char *)(IO_ADDRESS(TS7XXX_FLASH_BASE) + 0x7c030b); \
	(pMac)[2] = *(unsigned char *)(IO_ADDRESS(TS7XXX_FLASH_BASE) + 0x7c030c); \
	(pMac)[3] = *(unsigned char *)(IO_ADDRESS(TS7XXX_FLASH_BASE) + 0x7c030d); \
	(pMac)[4] = *(unsigned char *)(IO_ADDRESS(TS7XXX_FLASH_BASE) + 0x7c030e); \
	(pMac)[5] = *(unsigned char *)(IO_ADDRESS(TS7XXX_FLASH_BASE) + 0x7c030f); \
} \
}
	
/****  default MAC address if GET_MAC_ADDR does not override  *************/
static const U8  default_mac[6] = {0x00, 0xba, 0xd0, 0x0b, 0xad, 0x00};

/*****************************************************************************
* waitOnReg32()
*****************************************************************************/
static int  waitOnReg32(struct Ctlr *pD, int reg,
                        unsigned long mask, unsigned long expect, int tout)
{
    int  i;
    int  dt = 0;

/**/    _PRTK_ENTRY(("waitOnReg32(pD:%p,reg:0x%x,mask:0x%lux,expect:0x%lux,tout:%d)\n",
                 pD,reg,mask,expect,tout));
/**/

    for(i=0;i<10000;) {
        dt=RegRd32(reg);
        dt=(dt^expect)&mask;
        if(0==dt)  break;
        if(tout)  i++;
    }/*for*/
	_PRTK_INFO(("waitOnReg32 returns 0x%x\n", dt));
    return dt;
}/*waitOnReg32()*/

/*****************************************************************************
* eth_rxCtl()
*****************************************************************************/
static int  eth_rxCtl(struct Ctlr *pD,int sw) {

    _PRTK_ENTRY(("eth_rxCtl(pD:0x%p,sw:%d)\n",pD,sw));

/* Workaround for MAC lost 60-byte-long frames:  must enable Runt_CRC_Accept bit */
    RegWr32(REG_RxCTL, sw?RegRd32(REG_RxCTL)| RxCTL_SRxON | RxCTL_RCRCA:
                         RegRd32(REG_RxCTL)&~RxCTL_SRxON);

    return 0;
}/*eth_rxCtl()*/

/*****************************************************************************
* eth_indAddrWr()
*****************************************************************************/
static int  eth_indAddrWr(struct Ctlr *pD,int afp,char *pB) {
	unsigned char *pBuf = (unsigned char *) pB;
    U32  rxctl;
    int  i,  len;

    _PRTK_ENTRY(("eth_indAddrWr(pD:%p,afp:0x%x,pBuf:%p)\n",pD,afp, pBuf));
//	_PRTK_INFO(("NOT DOING IT\n"));
//	return 0;
    afp&=0x07;
    if(4==afp || 5==afp) {
        _PRTK_SWERR(("eth_indAddrWr(): invalid afp value\n"));
        return -1;
    }/*if*/
    len=(AFP_AFP_HASH==afp)?8:6;

    _PRTK_INFO(("eth_indAddrWr(): %02ux:%02ux:%02ux:%02ux:%02ux:%02ux%s",
                pBuf[0],pBuf[1],pBuf[2],pBuf[3],pBuf[4],pBuf[5],(6==len)?"\n":":"));
    if(6!=len) {
        _PRTK_INFO(("%02ux:%02ux\n",pBuf[6],pBuf[7]));
    }/*if*/

    rxctl=RegRd32(REG_RxCTL);    /*turn Rx off*/
    RegWr32(REG_RxCTL,~RxCTL_SRxON&rxctl);
    RegWr32(REG_AFP,afp);  /*load new address pattern*/
    for(i=0;i<len;i++)  RegWr8(REG_IndAD+i,pBuf[i]);
    RegWr32(REG_RxCTL,rxctl);  /*turn Rx back*/

    return 0;
}/*eth_indAddrWr()*/


#define  phy_wr(reg,dt)  _phy_write(pD, pD->priv.s.miiIdPhy, (reg), (dt))
#define  phy_rd(reg)     _phy_read(pD, pD->priv.s.miiIdPhy, (reg))
#define  phy_waitRdy()  waitOnReg32(pD, REG_MIISts,MIISts_Busy, ~MIISts_Busy, 1)

/*****************************************************************************
* _phy_write()
*****************************************************************************/
static void  _phy_write(struct Ctlr *pD,int idPhy,int reg,U16 dt) {

/*    _PRTK_ENTRY(("_phy_write(pD:0x%p,idPhy:%d,reg:0x%x,dt:0x%x)\n",pD,idPhy,reg,dt));*/

    phy_waitRdy();
    RegWr32(REG_MIIData,dt);
    RegWr32(REG_MIICmd,MIICmd_OP_WR|((idPhy&0x1f)<<5)|((reg&0x1f)<<0));
}/*_phy_write()*/



/*****************************************************************************
* _phy_read()
*****************************************************************************/
static U16  _phy_read(struct Ctlr *pD,int idPhy,int reg) {
    U16  dt;

    _PRTK_ENTRY(("_phy_read(pD:0x%p,idPhy:%d,reg:0x%x)\n",pD,idPhy,reg)); 

    phy_waitRdy();
    RegWr32(REG_MIICmd,MIICmd_OP_RD|((idPhy&0x1f)<<5)|((reg&0x1f)<<0));
    phy_waitRdy();
    dt=(unsigned short)RegRd32(REG_MIIData);

    _PRTK_INFO(("_phy_read(): read 0x%x\n",dt));/**/

    return dt;
}/*_phy_read()*/



#ifndef _PRTK_DUMP
#define _dbg_phy_dumpReg(pD)
#else
/*****************************************************************************
* _dbg_phy_dumpReg()
*****************************************************************************/
static void  _dbg_phy_dumpReg(struct Ctlr *pD) {
    _PRTK_DUMP(("Dumping registers of Ethernet PHY\n"));
    _PRTK_DUMP((" pD:0x%p, Eth Base Address:%p\n", pD,pD->base_addr));

    _PRTK_DUMP((" 0-3:0x%04x 0x%04x  0x%04x 0x%04x\n",
                phy_rd(0),phy_rd(1),phy_rd(2),phy_rd(3)));
    _PRTK_DUMP((" 4-6:0x%04x 0x%04x  0x%04x\n",
                phy_rd(4),phy_rd(5),phy_rd(6)));
    _PRTK_DUMP((" 16-19:0x%04x 0x%04x  0x%04x  0x%04x\n",
                phy_rd(16),phy_rd(17),phy_rd(18),phy_rd(19)));
    _PRTK_DUMP((" 20:0x%04x\n",phy_rd(20)));
}/*_dbg_phy_dumpReg()*/
#endif/*ifndef _PRTK_DUMP else*/


/*****************************************************************************
* phy_autoNegotiation()
*****************************************************************************/
static int  phy_autoNegotiation(struct Ctlr *pD) {
   U16 val = 0;
   U16 oldVal = 0;
	USED(oldVal, val, pD);
// XXX Already init'ed from RedBoot, don't waste time renegotiating
#define X
#ifdef X
   phy_wr(4, 0x01e1);  /*Set 802.3, 100M/10M Full/Half ability*/
   phy_wr(0,(1<<12)|(1<<9)); /* enable auto negotiation*/
   while (1) {
     val = phy_rd(1); /* read BM status Reg*/
     if ( val&0x0020) /* if Auto_Neg_complete?*/
     {
       break;
     }
   }

   //PHY needs the delay.  Otherwise it won't send the 1st frame.
   delay(1000);//delay 1 second.
#endif
   val = phy_rd(5); /* read ANLPAR Reg*/
	_PRTK_INFO(("phy_autoNegotiation: val of ANLPAR is 0x%ux\n", val));
   if ( val&0x0140) /* if 100M_FDX or 10M_FDX?*/
   {
     oldVal=RegRd16(REG_TestCTL);
     /*Enable MAC's Full Duplex mode.*/
     RegWr16(REG_TestCTL, oldVal|TestCTL_MFDX); 
     _PRTK_INFO(("ep93xx_eth(): set MAC to Full Duplex mode\n"));
   } else {
     _PRTK_INFO(("ep93xx_eth(): set MAC to Half Duplex mode\n"));
   }
   // TODO PJJ-added to get rid of _dbg_phy_dumpReg be sure returning zero is ok.


   return(0);
}

/*****************************************************************************
* phy_init()
*****************************************************************************/
static int  phy_init(struct Ctlr *pD) {
    U32 oldVal;

    _PRTK_ENTRY(("phy_init(pD:%p)\n",pD));

    oldVal = RegRd32(REG_SelfCTL);
    RegWr32(REG_SelfCTL, 0x6320); //0x6200);/*Set MDC clock to be divided by 32
                              // and disable PreambleSuppress bit*/
    phy_autoNegotiation(pD);

    RegWr32(REG_SelfCTL, oldVal);/*restore the old value */

    return 0;
}/*phy_init()*/

#ifndef _PRTK_DUMP
# define _dbg_ep9312Eth_dumpQueues(pD)
#else
/*****************************************************************************
* _dbg_ep9312Eth_dumpQueues()
*****************************************************************************/
static void  _dbg_ep9312Eth_dumpQueues(struct Ctlr *pD) {
    struct ep9213Eth_info  *pP=&pD->priv;
    int  i;

    _PRTK_DUMP(("Dumping Descriptor/Status Queues\n"));
    _PRTK_DUMP((" pD:0x%p, Base Address:%p\n",pD,pD->base_addr));

    _PRTK_DUMP((" o Rx Status Queue: at 0x%p, %d entries\n",pP->s.pQueRxSts,LEN_QueRxSts));
    for(i=0;i<LEN_QueRxSts;i++)
        _PRTK_DUMP(("  - %2d: 0x%08lux%08lux \n",
                    i,pP->s.pQueRxSts[i].e1,pP->s.pQueRxSts[i].e0));

    _PRTK_DUMP((" o Rx Descriptor Queue: at 0x%p, %d entries\n",pP->s.pQueRxDesc,LEN_QueRxDesc));
    for(i=0;i<LEN_QueRxDesc;i++)
        _PRTK_DUMP(("  - %2d: 0x%08lux%08lux \n",
                    i,pP->s.pQueRxDesc[i].e1, pP->s.pQueRxDesc[i].e0));

    _PRTK_DUMP((" o Tx Status Queue: at 0x%p, %d entries\n",pP->s.pQueTxSts,LEN_QueTxSts));
    for(i=0;i<LEN_QueTxSts;i++)
        _PRTK_DUMP(("  - %2d: 0x%08lux \n",i,pP->s.pQueTxSts[i].e0));

    _PRTK_DUMP((" o Tx Descriptor Queue: at 0x%p, %d entries\n",pP->s.pQueTxDesc,LEN_QueTxDesc));
    for(i=0;i<LEN_QueTxDesc;i++)
        _PRTK_DUMP(("  - %2d: 0x%08lux%08lux \n",
                    i, pP->s.pQueTxDesc[i].e1, pP->s.pQueTxDesc[i].e0));
}/*_dbg_ep9312Eth_dumpQueues()*/
#endif/*ifndef _PRTK_DUMP else*/

static unsigned char gTxDataBuff[LEN_QueTxDesc][LEN_TxBuf];
/*****************************************************************************
* devQue_start()
*
  make descriptor queues active
  allocate queue entries if needed
  and set device registers up to make it operational
  assume device has been initialized
* 
*****************************************************************************/
static int  devQue_start(struct Ctlr *pD) {
    int  err;
    struct ep9213Eth_info  *pP=&pD->priv;
    int  i;
    unsigned char  *pBuf;
    U32  phyA;

    _PRTK_ENTRY(("devQue_start(pD:0x%p)\n",pD));

    /*turn off device bus mastering*/
    RegWr32(REG_BMCtl,BMCtl_RxDis|BMCtl_TxDis|RegRd32(REG_BMCtl));
    err=waitOnReg32(pD,REG_BMSts,BMSts_TxAct,~BMSts_TxAct,1);
    err|=waitOnReg32(pD,REG_BMSts,BMSts_RxAct,~BMSts_RxAct,1);
    if(err) {
        _PRTK_HWFAIL(("devQue_start(): BM does not stop\n"));
    }/*if*/

    /*Tx Status Queue*/
    memset(pP->s.pQueTxSts, 0, sizeof(pP->s.pQueTxSts[0])*LEN_QueTxSts);  /*clear status*/
    pP->d.idxQueTxSts=0;  /*init index*/
    RegWr32(REG_TxSBA,pP->s.phyQueTxSts);  /*base addr*/
    RegWr32(REG_TxSCA,pP->s.phyQueTxSts);  /*current addr*/
    RegWr16(REG_TxSBL,sizeof(pP->s.pQueTxSts[0])*LEN_QueTxSts);  /*base len*/
    RegWr16(REG_TxSCL,sizeof(pP->s.pQueTxSts[0])*LEN_QueTxSts);  /*current len*/

    /*Tx Descriptor Queue*/
    memset(pP->s.pQueTxDesc, 0, sizeof(pP->s.pQueTxDesc[0])*LEN_QueTxDesc); /*clear descriptor*/
    pP->d.idxQueTxDescHead=pP->d.idxQueTxDescTail=0;  /*init index*/
    RegWr32(REG_TxDBA, pP->s.phyQueTxDesc);  /*base addr*/
    RegWr32(REG_TxDCA, pP->s.phyQueTxDesc);  /*current addr*/
    RegWr16(REG_TxDBL, sizeof(pP->s.pQueTxDesc[0])*LEN_QueTxDesc);  /*base len*/
    RegWr16(REG_TxDCL, sizeof(pP->s.pQueTxDesc[0])*LEN_QueTxDesc);  /*current len*/

    /*Rx Status Queue*/
    memset(pP->s.pQueRxSts, 0, sizeof(pP->s.pQueRxSts[0])*LEN_QueRxSts);  /*clear status*/
    pP->d.idxQueRxSts=0;  /*init index*/
    RegWr32(REG_RxSBA, pP->s.phyQueRxSts);  /*base addr*/
    RegWr32(REG_RxSCA, pP->s.phyQueRxSts);  /*current addr*/
    RegWr16(REG_RxSBL, sizeof(pP->s.pQueRxSts[0])*LEN_QueRxSts);  /*base len*/
    RegWr16(REG_RxSCL, sizeof(pP->s.pQueRxSts[0])*LEN_QueRxSts);  /*current len*/

    /*Rx Descriptor Queue*/
    memset(pP->s.pQueRxDesc, 0, sizeof(pP->s.pQueRxDesc[0])*LEN_QueRxDesc);  /*clear descriptor*/
    phyA=pP->s.phyRxBuf;
    for(i=0;i<LEN_QueRxDesc;i++) {
        pP->s.pQueRxDesc[i].e1=i<<16;  /*Rx Buffer Index*/
        pP->s.pQueRxDesc[i].e0=phyA;  /*physical address of Rx Buf*/
        pP->s.pQueRxDesc[i].e1 |= LEN_RxBuf;  /*Rx Buffer Length*/
        phyA+=(LEN_RxBuf+3)&~0x03;
    }/*for*/


    pP->d.idxQueRxDesc=0;  /*init index*/
    RegWr32(REG_RxDBA, pP->s.phyQueRxDesc);  /*base addr*/
    RegWr32(REG_RxDCA, pP->s.phyQueRxDesc);  /*current addr*/
    RegWr16(REG_RxDBL, sizeof(pP->s.pQueRxDesc[0])*LEN_QueRxDesc);  /*base len*/
    RegWr16(REG_RxDCL, sizeof(pP->s.pQueRxDesc[0])*LEN_QueRxDesc);  /*current len*/

    /*init Rx Buffer Descriptors*/
    pBuf=pP->s.pRxBuf;
    for(i=0;i<LEN_QueRxDesc;i++) {
        pP->s.pRxBufDesc[i].vaddr=pBuf;  /*system address of buffer*/
        pBuf+=(LEN_RxBuf+3)&~0x03;
    }/*for*/
	/* IT IS OK TO THIS POINT. */
    /*init Tx Buffer Descriptors*/


    memset(pP->s.pTxBufDesc, 0x0, sizeof(*pP->s.pTxBufDesc)*LEN_QueTxDesc);
//return 0;
	pP->s.pTxBuf = &gTxDataBuff[0][0];
    for(i=0;i<LEN_QueTxDesc;i++) {

        pP->s.pTxBufDesc[i].vaddr=&gTxDataBuff[i][0];  /*system address of buffer*/
		iprint("do set %p to %p\n", 
			&pP->s.pTxBufDesc[i].vaddr, 
			pP->s.pTxBufDesc[i].vaddr);
    }/*for*/
//print("NOT enabling hardware ...\n"); return 0;
    /*turn on device bus mastering*/
    RegWr32(REG_BMCtl, BMCtl_TxEn|BMCtl_RxEn|RegRd32(REG_BMCtl));
    err=waitOnReg32(pD, REG_BMSts,BMSts_TxAct|BMSts_TxAct,BMSts_TxAct|BMSts_TxAct,1);
    if(err) {
        _PRTK_HWFAIL(("devQue_start(): BM does not start\n"));
    }/*if*/

    /*Enqueue whole entries; this must be done after BM activation*/
    RegWr32(REG_RxSEQ, LEN_QueRxSts);  /*Rx Status Queue*/
    RegWr32(REG_RxDEQ, LEN_QueRxDesc);  /*Rx Descriptor queue*/

    return 0;
}/*devQue_start()*/



/*****************************************************************************
* devQue_init()
  init device descriptor queues at system level
  device access is not recommended at this point
* 
*****************************************************************************/
static int  devQue_init(struct Ctlr *pD) {
    struct ep9213Eth_info  *pP=&pD->priv;
    unsigned char  *pBuf;
    int   size;

    _PRTK_ENTRY(("devQue_init(pD:0x%x)\n", (unsigned int)pD));

    /*verify device Tx/Rx Descriptor/Status Queue data size*/
    if(8 != sizeof(receiveDescriptor)) {
        _PRTK_SWERR(("devQue_init(): size of receiveDescriptor is not 8 bytes!!!\n"));
        return -1;
    }else if(8 != sizeof(receiveStatus)) {
        _PRTK_SWERR(("devQue_init(): size of receiveStatus is not 8 bytes!!!\n"));
        return -1;
    }else if(8 != sizeof(transmitDescriptor)) {
        _PRTK_SWERR(("devQue_init(): size of transmitDescriptor is not 8 bytes!!!\n"));
        return -1;
    }else if(4 != sizeof(transmitStatus)) {
        _PRTK_SWERR(("devQue_init(): size of transmitStatus is not 4 bytes!!!\n"));
        return -1;
    }/*else*/

    /*
      allocate kernel memory for whole queues
      best if non-cached memory block due to DMA access by the device
      if CPU doesn't have bus snooping
    */
    size=sizeof(receiveDescriptor)*(LEN_QueRxDesc+1)+
         sizeof(receiveStatus)*(LEN_QueRxSts+1)+
         sizeof(transmitDescriptor)*(LEN_QueTxDesc+1)+
         sizeof(transmitStatus)*(LEN_QueTxSts+1)+
         sizeof(unsigned long)*4;

    pBuf=MALLOC_DMA(size, &pP->s.phyQueueBase);  /*request memory allocation*/
    if(!pBuf) {
        _PRTK_SYSFAIL(("devQue_initAll(): fail to allocate memory for queues\n"));
        return -1;
    }/*if*/
	_PRTK_INFO(("devQue_initAll():pBuf %p, pP->s.phyQueueBase 0x%lux\n", pBuf, pP->s.phyQueueBase));
    /*
      assign memory to each queue
    */
    pP->s.pQueRxDesc = (void*)Align32(pBuf);
    pBuf = (unsigned char*)pBuf + sizeof(receiveDescriptor)*(LEN_QueRxDesc+1);
    pP->s.pQueRxSts = (void*)Align32(pBuf);
    pBuf = (unsigned char*)pBuf + sizeof(receiveStatus)*(LEN_QueRxSts+1);
    pP->s.pQueTxDesc = (void*)Align32(pBuf);
    pBuf = (unsigned char*)pBuf + sizeof(transmitDescriptor)*(LEN_QueTxDesc+1);
    pP->s.pQueTxSts = (void*)Align32(pBuf);
//    pBuf = (unsigned char*)pBuf + sizeof(transmitStatus)*(LEN_QueTxSts+1);

    /*
      store physical address of each queue
    */
    pP->s.phyQueRxDesc = Align32(pP->s.phyQueueBase);
    pP->s.phyQueRxSts = pP->s.phyQueRxDesc + ((U32)pP->s.pQueRxSts-(U32)pP->s.pQueRxDesc);
    pP->s.phyQueTxDesc = pP->s.phyQueRxDesc + ((U32)pP->s.pQueTxDesc-(U32)pP->s.pQueRxDesc);
    pP->s.phyQueTxSts = pP->s.phyQueRxDesc + ((U32)pP->s.pQueTxSts-(U32)pP->s.pQueRxDesc);

    _PRTK_INFO(("devQue_init(): Rx Descriptor Queue at 0x%p(Phy0x%x), %d entries\n",
                pP->s.pQueRxDesc, (unsigned int)pP->s.phyQueRxDesc,LEN_QueRxDesc));
    _PRTK_INFO(("devQue_init(): Rx Status Queue at 0x%p(Phy0x%x), %d entries\n",
                pP->s.pQueRxSts, (unsigned int)pP->s.phyQueRxSts,LEN_QueRxSts));
    _PRTK_INFO(("devQue_init(): Tx Descriptor Queue at 0x%p(Phy0x%x), %d entries\n",
                pP->s.pQueTxDesc, (unsigned int)pP->s.phyQueTxDesc,LEN_QueTxDesc));
    _PRTK_INFO(("devQue_init(): Tx Status Queue at 0x%p(Phy0x%x), %d entries\n",
                pP->s.pQueTxSts, (unsigned int)pP->s.phyQueTxSts,LEN_QueTxSts));

    /*
      init queue entries
    */
    memset(pP->s.pQueRxDesc, 0, sizeof(receiveDescriptor)*LEN_QueRxDesc);
    memset(pP->s.pQueRxSts, 0, sizeof(receiveStatus)*LEN_QueRxSts);
    memset(pP->s.pQueTxDesc, 0, sizeof(transmitDescriptor)*LEN_QueTxDesc);
    memset(pP->s.pQueTxSts, 0, sizeof(transmitStatus)*LEN_QueTxSts);

    /*Allocate Rx Buffer
      (We might need to copy from Rx buf to skbuff in whatever case,
       because device bus master requires 32bit aligned Rx buffer address 
       but Linux network stack requires odd 16bit aligned Rx buf address)*/
    pP->s.pRxBuf = MALLOC_DMA(((LEN_RxBuf+3)&~0x03)*LEN_QueRxDesc, &pP->s.phyRxBuf);
    if(!pP->s.pRxBuf) { 
        /*@*/
        pP->s.pRxBuf=0;
        _PRTK_SYSFAIL(("devQue_init(): fail to allocate memory for RxBuf\n"));
        return -1;
    }/*if*/


    /*
      allocate kernel memory for buffer descriptors
    */
    size = sizeof(bufferDescriptor)*(LEN_QueRxDesc+LEN_QueTxDesc);
    pBuf = malloc(size);
//	pBuf = MALLOC_DMA(size, 0);
    if(!pBuf) {
        _PRTK_SYSFAIL(("devQue_initAll(): fail to allocate memory for buf desc\n"));
        return -1;
    }/*if*/
    memset(pBuf,0x0,size);  /*clear with 0*/
    pP->s.pRxBufDesc= (bufferDescriptor *) pBuf;
    pP->s.pTxBufDesc= (bufferDescriptor *) (pBuf+sizeof(bufferDescriptor)*LEN_QueRxDesc);
	_PRTK_INFO(("pRxBufDesc %p pTxBufDesc %p\n", 
			pP->s.pRxBufDesc, pP->s.pTxBufDesc));
    return 0;
}/*devQue_init()*/



/*****************************************************************************
* eth_reset()
*****************************************************************************/
static int  eth_reset(struct Ctlr *pD) {
    int  err;

    _PRTK_ENTRY(("eth_reset(pD:0x%p)\n",pD));

    RegWr8(REG_SelfCTL,SelfCTL_RESET);  /*soft reset command*/
    err=waitOnReg32(pD,REG_SelfCTL,SelfCTL_RESET,~SelfCTL_RESET,1);
    if(err) {
        _PRTK_WARN(("eth_reset(): Soft Reset does not self-clear\n"));
    }/*if*/

    /*reset PHY*/
    //phy_reset(pD);

    return 0;
}/*eth_reset()*/





/*****************************************************************************
*  eth_enable()
 
  Purpose:
        Turn on device interrupt for interrupt driven operation.
        Also turn on Rx but no Tx.
* 
*****************************************************************************/
static int  eth_enable(struct Ctlr *pD) {

    _PRTK_ENTRY(("eth_enable(pD:0x%p)\n",pD));

    RegWr32(REG_IntEn,Default_IntSrc);  /*setup Interrupt sources*/
    RegWr32(REG_GIntMsk,GIntMsk_IntEn);  /*turn on INT*/
    eth_rxCtl(pD,1);  /*turn on Rx*/

    return 0;
}/*eth_enable()*/


/*****************************************************************************
*  eth_init()
 
  Purpose:
        Reset and initialize the device.
        Device should be initialized enough to function in polling mode.
        Tx and Rx must be disabled and no INT generation.
* 
*****************************************************************************/
static int  eth_init(struct Ctlr *pD) {
	int i;

    _PRTK_ENTRY(("eth_init(pD:0x%p), base addr %p\n",pD, pD->base_addr));

    /*reset device*/
    eth_reset(pD);

	/*init PHY*/
    phy_init(pD);

    /*init MAC*/
    RegWr32(REG_SelfCTL,0x6300);/*Set MDC clock to be divided by 32
                               and enable PreambleSuppress bit*/
    RegWr32(REG_GIntMsk,0x00);  /*mask Interrupt*/
    RegWr32(REG_RxCTL,RxCTL_BA|RxCTL_IA0);  /*no Rx on at this point*/
    RegWr32(REG_TxCTL,0x00);
    RegWr32(REG_GT,0x00);
    RegWr32(REG_BMCtl,0x00);
    RegWr32(REG_RxBTH,(0x80<<16)|(0x40<<0));  /*Buffer Threshold*/
    RegWr32(REG_TxBTH,(0x80<<16)|(0x40<<0) );
    RegWr32(REG_RxSTH,(4<<16)|(2<<0) );        /*Status Threshold*/
    RegWr32(REG_TxSTH,(4<<16)|(2<<0) );
    RegWr32(REG_RxDTH,(4<<16)|(2<<0) );        /*Descriptor Threshold*/
    RegWr32(REG_TxDTH,(4<<16)|(2<<0) );
    RegWr32(REG_MaxFL,((1518+1)<<16)|(944<<0));  /*Max Frame Length & Tx Start Threshold*/
    
    RegRd32(REG_TxCollCnt);  /*clear Tx Collision Counter*/
    RegRd32(REG_RxMissCnt);  /*clear Rx Miss Counter*/
    RegRd32(REG_RxRntCnt);   /*clear Rx Runt Counter*/

    RegRd32(REG_IntStsC);  /*clear Pending INT*/

    RegWr32(REG_TxCTL,TxCTL_STxON|RegRd32(REG_TxCTL));  /*Tx on*/
    /*Get the Ethernet MAC address from system
      but, do NOT set to the device at this point*/
    for(i=0;i<6;i++)  pD->dev_addr[i]=default_mac[i];
    GET_MAC_ADDR(pD,&pD->dev_addr[0]);

    /*Set MAC address*/
    eth_indAddrWr(pD,AFP_AFP_IA0,&pD->dev_addr[0]);

    /*init queue*/
    devQue_start(pD);

    return 0;
}/*eth_init()*/


/* END of GPL'ed code from arm *******************/
enum {					/* registers */
	Cr		= 0x37,		/* e.g. Command Register */
	Imr		= 0x3C,		/* e.g. Interrupt Mask */
	Isr		= 0x3E,		/* e.g. Interrupt Status */
	Tcr		= 0x40,		/* e.g. Transmit Configuration */
	Rcr		= 0x44,		/* e.g. Receive Configuration */
	/*
	 * Change the examples to suit and fill in the rest.
	 */
};

enum {					/* Cr */
	Te		= 0x04,		/* e.g. Transmitter Enable */
	/*
	 * Change the examples to suit and fill in the rest.
	 */
};

enum {					/* Imr/Isr */
	Rok		= 0x0001,	/* Receive OK */
	Rer		= 0x0002,	/* Receive Error */
	Tok		= 0x0004,	/* Transmit OK */
	Ter		= 0x0008,	/* Transmit Error */
	Rdu		= 0x0010,	/* Receive Descriptor Unavailable */
	Punlc		= 0x0020,	/* Packet Underrun or Link Change */
	Fovw		= 0x0040,	/* Receive FIFO Overflow */
	Tdu		= 0x0080,	/* Transmit Descriptor Unavailable */
	Swint		= 0x0100,	/* Software Interrupt */
	Timeout		= 0x4000,	/* Timer */
	Serr		= 0x8000,	/* System Error */
	/*
	 * Change the examples to suit and fill in the rest.
	 */
};

enum {					/* Tcr */
	Acrc		= 0x00010000,	/* e.g. Append CRC (not) */
	/*
	 * Change the examples to suit and fill in the rest.
	 */
};

enum {					/* Rcr */
	Ab		= 0x00000008,	/* e.g. Accept Broadcast */
	/*
	 * Change the examples to suit and fill in the rest.
	 */
};

enum {					/* Transmit Descriptor control */
	TxflMASK	= 0x0000FFFF,	/* e.g. Transmit Frame Length */
	TxflSHIFT	= 0,
	/*
	 * Change the examples to suit and fill in the rest.
	 */
};

enum {					/* Receive Descriptor control */
	RxflMASK	= 0x00003FFF,	/* e.g. Receive Frame Length */
	RxflSHIFT	= 0,
	Tcpf		= 0x00004000,	/* TCP Checksum Failure */
	Udpf		= 0x00008000,	/* UDP Checksum Failure */
	Ipf		= 0x00010000,	/* IP Checksum Failure */
	Pid0		= 0x00020000,	/* Protocol ID0 */
	Pid1		= 0x00040000,	/* Protocol ID1 */
	Crce		= 0x00080000,	/* CRC Error */
	Runt		= 0x00100000,	/* Runt Packet */
	Res		= 0x00200000,	/* Receive Error Summary */
	Rwt		= 0x00400000,	/* Receive Watchdog Timer Expired */
	Fovf		= 0x00800000,	/* FIFO Overflow */
	Bovf		= 0x01000000,	/* Buffer Overflow */
	Bar		= 0x02000000,	/* Broadcast Address Received */
	Pam		= 0x04000000,	/* Physical Address Matched */
	Mar		= 0x08000000,	/* Multicast Address Received */
	/*
	 * Change the examples to suit and fill in the rest.
	 */
};

enum {					/* General Descriptor control */
	Ls		= 0x10000000,	/* e.g. Last Segment Descriptor */
	Fs		= 0x20000000,	/* e.g. First Segment Descriptor */
	Eor		= 0x40000000,	/* e.g. End of Descriptor Ring */
	Own		= 0x80000000,	/* e.g. Ownership */
};

/*
 */
enum {					/* Ring sizes  (<= 1024) */
	Ntd		= 128,		/* Transmit Ring */
	Nrd		= 64,		/* Receive Ring */

	Mps		= ROUNDUP(ETHERMAXTU+4, 128),
};


static Ctlr* ctlrhead;
static Ctlr* ctlrtail;

#define csr8r(c, r)	(inb((c)->port+(r)))
#define csr16r(c, r)	(ins((c)->port+(r)))
#define csr32r(c, r)	(inl((c)->port+(r)))
#define csr8w(c, r, b)	(outb((c)->port+(r), (int)(b)))
#define csr16w(c, r, w)	(outs((c)->port+(r), (ushort)(w)))
#define csr32w(c, r, l)	(outl((c)->port+(r), (ulong)(l)))
#ifdef NO
static int
ep93xxmiimir(Mii* mii, int pa, int ra)
{
	/*
	 * Given a PHY address (pa) and register within
	 * that PHY (ra), return the 16-bit contents or
	 * -1 on error.
	 */
	USED(mii, pa, ra);

	return -1;
}

static int
ep93xxmiimiw(Mii* mii, int pa, int ra, int data)
{
	/*
	 * Given a PHY address (pa) and register within
	 * that PHY (ra), write the 16-bit contents with data or
	 * -1 on error. Return 0 on success.
	 */
	USED(mii, pa, ra, data);

	return -1;
}

static int
ep93xxmii(Ctlr* ctlr)
{
	MiiPhy *phy;

	/*
	 * Link management.
	 */
	if((ctlr->mii = malloc(sizeof(Mii))) == nil)
		return -1;
	ctlr->mii->mir = ep93xxmiimir;
	ctlr->mii->miw = ep93xxmiimiw;
	ctlr->mii->ctlr = ctlr;

	if(mii(ctlr->mii, ~0) == 0 || (phy = ctlr->mii->curphy) == nil){
		free(ctlr->mii);
		ctlr->mii = nil;
		return -1;
	}
	print("oui %X phyno %d\n", phy->oui, phy->phyno);

	miiane(ctlr->mii, ~0, ~0, ~0);

	return 0;
}
#endif

static void
ep93xxpromiscuous(void* arg, int on)
{
	Ether *edev;
	Ctlr * ctlr;

	edev = arg;
	ctlr = edev->ctlr;
	ilock(&ctlr->ilock);

	if(on){
		/*
		 * Set any bits to turn promiscuous on here.
		 */
		;
	}
	else{
		/*
		 * Clear any bits to turn off promiscuous here.
		 */
		;
	}
	iunlock(&ctlr->ilock);
}

static long
ep93xxifstat(Ether* edev, void* a, long n, ulong offset)
{
	char *p;
	Ctlr *ctlr;
	int i, l, r;
	SET(i,r);
	USED(i,r);
	ctlr = edev->ctlr;
	qlock(&ctlr->slock);

	p = nil;
	if(waserror()){
		qunlock(&ctlr->slock);
		free(p);
		nexterror();
	}

	/*
	 * Do any special chip commands to extract statistics here.
	 * and update edev->[oerrs|crcs|frames|buffs|overflows]
	 * if the receive/transmit interrupt routines can't do it.
	 * If n == 0, just wanted statistics updates.
	 */

	if(n == 0){
		qunlock(&ctlr->slock);
		poperror();
		return 0;
	}

	if((p = malloc(READSTR)) == nil)
		error(Enomem);

	l = 0;
	USED(l);
	l += snprint(p+l, READSTR-l, "ierrs: %ud\n", ctlr->ierrs);
/*
	l = snprint(p+l, READSTR-l, "txdu: %ud\n", ctlr->txdu);
	l += snprint(p+l, READSTR-l, "tcpf: %ud\n", ctlr->tcpf);
	l += snprint(p+l, READSTR-l, "udpf: %ud\n", ctlr->udpf);
	l += snprint(p+l, READSTR-l, "ipf: %ud\n", ctlr->ipf);
	l += snprint(p+l, READSTR-l, "fovf: %ud\n", ctlr->fovf);
	l += snprint(p+l, READSTR-l, "ierrs: %ud\n", ctlr->ierrs);
	l += snprint(p+l, READSTR-l, "rer: %ud\n", ctlr->rer);
	l += snprint(p+l, READSTR-l, "rdu: %ud\n", ctlr->rdu);
	l += snprint(p+l, READSTR-l, "punlc: %ud\n", ctlr->punlc);
	l += snprint(p+l, READSTR-l, "fovw: %ud\n", ctlr->fovw);

	l += snprint(p+l, READSTR-l, "tcr: %8.8uX\n", ctlr->tcr);
	l += snprint(p+l, READSTR-l, "rcr: %8.8uX\n", ctlr->rcr);
*/
/*
	if(ctlr->mii != nil && ctlr->mii->curphy != nil){
		l += snprint(p+l, READSTR, "phy:   ");
		for(i = 0; i < NMiiPhyr; i++){
			if(i && ((i & 0x07) == 0))
				l += snprint(p+l, READSTR-l, "\n       ");
			r = miimir(ctlr->mii, i);
			l += snprint(p+l, READSTR-l, " %4.4uX", r);
		}
		snprint(p+l, READSTR-l, "\n");
	}
*/
	n = readstr(offset, a, n, p);

	qunlock(&ctlr->slock);
	poperror();
	free(p);

	return n;
}

static int
ep93xxreset(Ctlr* ctlr)
{
	int timeo;

	/*
	 * Soft reset the controller.
	 * Return 0 on success, -1 on error.
	 * Make sure you timeout, e.g.
	csr8w(ctlr, Cr, Rst);
	for(timeo = 0; timeo < 1000; timeo++){
		if(!(csr8r(ctlr, Cr) & Rst))
			return 0;
		delay(1);
	}
	 */
	USED(ctlr);
	SET(timeo);
	USED(timeo);

	return -1;
}

static void
ep93xxhalt(Ctlr* ctlr)
{
	/*
	 * Turn off receive/transmit and disable all interrupts.
	 */
	USED(ctlr);
}

static void
ep93xxreplenish(Ctlr* ctlr)
{
	D *d;
	int rdt;
	Block *bp;

	/*
	 * Replenish the receive ring.
	 */
	rdt = ctlr->rdt;
	while(NEXT(rdt, ctlr->nrd) != ctlr->rdh){
		d = &ctlr->rd[rdt];
		if(ctlr->rb[rdt] == nil){
			/*
			 * simple allocation for now
			 */
			bp = iallocb(Mps);
			if(bp == nil){
				iprint("no available buffers\n");
				break;
			}
			ctlr->rb[rdt] = bp;
			/*
			 * Fill in ncessary fields in new descriptor, e.g.
			d->addrlo = PCIWADDR(bp->rp);
			d->addrhi = 0;
			 */
		}
		coherence();
		/*
		 * Make the descriptor available, e.g.
		d->control |= Own|Mps;
		 */
		USED(d);
		rdt = NEXT(rdt, ctlr->nrd);
		ctlr->nrdfree++;
	}
	ctlr->rdt = rdt;
}

/* we did way too much already in pnp .... oh well. */
static void
ep93xxinit(Ether* edev)
{


	Ctlr *ctlr;

	ctlr = edev->ctlr;
	ilock(&ctlr->ilock);

//	ep93xxhalt(ctlr);
//	ep93xxreplenish(ctlr);

	/*
	 * Interrupts.
	 * Configure masks, etc.
	 */

	/*
	 * Initialise any statistics gathering.
	 */

	/*
	 * Set configuration.
	 * [RT]cr, ring base addresses, etc.
	 * and enable the sucker.
	 */

	iunlock(&ctlr->ilock);

	/*
	 * This might be done here or earlier in pnp.
	 * Depends on the hardware.
	ep93xxmii(ctlr);
	 */
}

static void
ep93xxattach(Ether* edev)
{
	Ctlr *ctlr;

	ctlr = edev->ctlr;
	iprint("ETHERATTACH================================>");
	qlock(&ctlr->alock);
	/*
	 * On first attach, allocate reources and initialise.
	 */
	ctlr->init = 0;
	if(ctlr->init == 0){
	    /*init device descriptor/status queues*/
	    devQue_init(ctlr);
		/* note: this sets MAC address in ctlr */
		/**/
		if (eth_init(ctlr)) {
			poststr("eth_init failed\n");
			qunlock(&ctlr->alock);
			return;
		}
		memmove(edev->ea, ctlr->dev_addr, sizeof(edev->ea));
		iprint("ea %x %x %x %x %x %x\n", 
				edev->ea[0],edev->ea[1], edev->ea[2], edev->ea[3],  edev->ea[4],edev->ea[5]);
		/**/
		/* guess we'll just enable interrupts here */
		eth_enable(ctlr);
		intrenable(IRQ, IRQENETMAC, ep93xxinterrupt, edev, "ether");
		ctlr->init = 1;
	}
	qunlock(&ctlr->alock);

	/*
	 * Should wait for link to be ready here.
	 */
	iprint("ALL DONE attach\n");
}

static void
ep93xxlink(Ether* edev)
{
	uint r;
	int limit;
	Ctlr *ctlr;
	iprint("ep93xxlinke\n");
	ctlr = edev->ctlr;

	/*
	 * Link change interrupt.
	 * What happens here depends on what that hardware does.
	 * Sometimes the hardware does all the work, then all that needs
	 * to be done is noting the speed and limits.
	 * Sometimes have to set registers depending on any new
	 * speed/duplex settings.
	 */
	SET(r, limit);
	USED(ctlr, r, limit);
}
/*****************************************************************************
* eth_isrRx()
*
*  Interrupt Service Routines
* 
*****************************************************************************/
static int  eth_isrRx(struct Ctlr *pD)
{
	ep9213Eth_info  *pP = &pD->priv;
	receiveStatus   *pQRxSts;
	int  idxQRxStsHead;  /*index of Rx Status Queue Head from device (next put point)*/
	int  idxSts;
	int  cntStsProcessed,  cntDescProcessed;
	char  *pDest;
	int  len;
	ulong  dt;

	SET(pDest);
	USED(pDest);
	_PRTK_ENTRY_ISR(("eth_isrRx(pD:%p)\n", pD));

	/*get Current Rx Status Queue pointer*/
	dt=RegRd32(REG_RxSCA);
	idxQRxStsHead=(dt-pP->s.phyQueRxSts)/sizeof(pP->s.pQueRxSts[0]);  /*convert to array index*/
	_PRTK_INFO_ISR(("eth_isrRx(): RxSCA:0x%lux -> idx:%d\n",dt,idxQRxStsHead));
	if(!(0<=idxQRxStsHead&&idxQRxStsHead<LEN_QueRxSts)) {
		_PRTK_HWFAIL(("eth_isrRx(): invalid REG_RxSCA:0x%lux idx:%d (phyQueRxSts:0x%lux Len:%d)\n",
		    dt,idxQRxStsHead,pP->s.phyQueRxSts,LEN_QueRxSts));
		/*@ do recover*/
		return -1;
	}/*if*/

	/*process Rx (limit to idxQRxStsHead due to cache)*/
	cntStsProcessed=cntDescProcessed=0;
	while(idxQRxStsHead!=pP->d.idxQueRxSts) {

		idxSts=pP->d.idxQueRxSts;
		pP->d.idxQueRxSts=IdxNext(pP->d.idxQueRxSts,LEN_QueRxSts);
		pQRxSts=&pP->s.pQueRxSts[idxSts];
		_PRTK_INFO_ISR(("eth_isrRx(): QRxSts[%d]:0x%lux 0x%lux\n",idxSts,(unsigned long)pQRxSts->e0, (unsigned long)pQRxSts->e1));
		if(!(pQRxSts->e0 & (1<<31))) {  /*empty?*/
			_PRTK_HWFAIL(("eth_isrRx(): QueRxSts[%d] is empty; Hd:%d\n",
			    idxSts,idxQRxStsHead));
			/*@ do recover*/
			return -1;
		}/*if*/

		/* still in while ... */
		pQRxSts->e0 &= ~(1<<31);  /*mark processed*/

		if(pQRxSts->e0 && (1<28)) {  /*buffer has data*/
			int bi = (pQRxSts->e1>>16)&0x7fff;

			if(bi==pP->d.idxQueRxDesc) {  /*check BI*/


				int ok = (pQRxSts->e0 & (1<<30)) &&
				(pQRxSts->e0 & (1<<29));
				pP->d.idxQueRxDesc=IdxNext(pP->d.idxQueRxDesc,LEN_QueRxDesc);  /*inc Desc index*/
				cntDescProcessed++;
				if(ok) {  /*received a frame without error*/
					len=pQRxSts->e1 & 0xffff;
					_PRTK_INFO_ISR(("eth_isrRx(): Rx Len:%d\n",len));
					{ /* queue bp */
						Block *bp;
						bp = iallocb(len);

						_PRTK_INFO_ISR(("eth_isRx(): copy to %p\n", 
						    bp->base));
						memmove(bp->wp, 
						    pP->s.pRxBufDesc[bi].vaddr, len);
/*
						{ int i, j;
							for(i = 0; i < len; i += 16){
								iprint("0x%x: ", i);
								for(j = 0; i + j < len; j ++)
									iprint("%02x ", bp->wp[i+j]);
								iprint("\n");
							}
						}
*/
						bp->wp += len;
						bp->next = nil;
						etheriq(pD->edev, bp, 1);
					} /* end queue bp */
			}/* else OK */ else {  /*errored Rx*/
				_PRTK_INFO_ISR(("eth_isrRx(): errored Rx, QueRxSts[%d]:0x%lux/0x%lux\n",
				    idxSts, pP->s.pQueRxSts[idxSts].e0,
					pP->s.pQueRxSts[idxSts].e1));
				// pP->d.stats.rx_errors++;
#ifdef NOTYET
				if(pQRxSts->f.oe)  pP->d.stats.rx_fifo_errors++;  /*overrun*/
				if(pQRxSts->f.fe)  pP->d.stats.rx_frame_errors++;  /*frame error*/
				if(pQRxSts->f.runt||pQRxSts->f.edata)  pP->d.stats.rx_length_errors++;  /*inv length*/
				if(pQRxSts->f.crce)  pP->d.stats.rx_crc_errors++;  /*crc error*/
#endif
					}  
			}/* bi mismatch */ else {
				_PRTK_HWFAIL(("eth_isrRx(): unmatching QueRxSts[%d].BI:0x%x; idxQueRxDesc:0x%x\n",
	    		idxSts,bi,pP->d.idxQueRxDesc));
			}/*else unmatching bi blowup*/

		}/*if*/ /* buffer has data */

		cntStsProcessed++;
	}/*while*/

	/*enqueue*/

	_PRTK_INFO_ISR(("eth_isrRx(): enqueue QueRxSts:%d QueRxDesc:%d\n",
	    cntStsProcessed,cntDescProcessed));
	RegWr32(REG_RxSEQ, cntStsProcessed);
	RegWr32(REG_RxDEQ, cntDescProcessed);
#endif

	return 0;
}/*eth_isrRx()*/






/*****************************************************************************
* ep9213Eth_isr()
*****************************************************************************/
static void ep9213Eth_isr(struct Ctlr *pD) {
    int  lpCnt;
    U32  intS;

    _PRTK_ENTRY_ISR(("ep9213Eth_isr(pDev:%p)\n",pD));
	if (! pD) {
		panic("NO PD in interrupt\n");
	}
    lpCnt=0;
    do {
        intS=RegRd32(REG_IntStsC);  /*get INT status and then clear*/

#ifdef NO
        intE=RegRd32(REG_IntEn);  /*get INT Source Enable*/
        intS&=(intE&~0x07)|(IntSts_AHBE|IntSts_OTHER|IntSts_SWI)|
              ((intE&0x07)?IntSts_RxSQ:0x00);
        _PRTK_INFO_ISR(("ep9213Eth_isr(): intS:0x%x intE:0x%x\n",intS,intE));
#endif

        if(!intS)  break;  /*no INT*/
        if(IntSts_RxSQ&intS)  eth_isrRx(pD);  /*Rx INT*/
        if(IntSts_TxSQ&intS)  eth_isrTx(pD);  /*Tx INT*/
    }while(lpCnt++ < 64);  /*limit loop to serve other interrupts too*/

}/*ep9213Eth_isr()*/

/*****************************************************************************
* eth_chkTxLvl()
*****************************************************************************/
static void  eth_chkTxLvl(struct Ctlr *pD) {
    struct ep9213Eth_info  *pP = &pD->priv;
    int  idxQTxDescHd;
    int  filled;

    _PRTK_ENTRY_ISR(("eth_chkTxLvl(pD:0x%x)\n", (unsigned int) pD));

    idxQTxDescHd=pP->d.idxQueTxDescHead;

    /*check Tx Descriptor Queue fill-up level*/
    filled=idxQTxDescHd-pP->d.idxQueTxDescTail;
    if(filled<0)  filled += LEN_QueTxDesc;

    if(pP->d.txStopped && filled <= (LVL_TxResume+1)) {
        pP->d.txStopped=0;
 //       pD->trans_start = jiffies;
//        netif_wake_queue(pD);
        _PRTK_INFO_ISR(("eth_chkTxLvl(): Tx Resumed, filled:%d\n",filled));
    }/*if*/
}/*eth_chkTxLvl()*/



/*****************************************************************************
* eth_cleanUpTx()
*****************************************************************************/
static int  eth_cleanUpTx(struct Ctlr *pD) {
    struct ep9213Eth_info  *pP = &pD->priv;
    transmitStatus  *pQTxSts;
    int  idxSts,  bi;

    _PRTK_ENTRY(("eth_cleanUpTx(pD:0x%x)\n", (unsigned int)pD));

    /*process Tx Status Queue (no need to limit processing of TxStatus Queue because
      each queue entry consist of 1 dword)*/
    while(pP->s.pQueTxSts[pP->d.idxQueTxSts].e0 & (1<<31)) {

        idxSts = pP->d.idxQueTxSts;

        /*ahead Tx Status Queue index*/
        pP->d.idxQueTxSts = IdxNext(pP->d.idxQueTxSts,LEN_QueTxSts);
        pQTxSts = &pP->s.pQueTxSts[idxSts];
        _PRTK_INFO(("eth_cleanUpTx(): QTxSts[%d]:0x%x\n",idxSts,(unsigned int)pQTxSts->e0));
        if(!(pQTxSts->e0&(1<<31))) {  /*empty?*/
            _PRTK_HWFAIL(("eth_cleanUpTx(): QueTxSts[%d]:x%08x is empty\n",idxSts,(int)pQTxSts->e0));
            /*@ do recover*/
            return -1;
        }/*if*/
        pQTxSts->e0 &= ~(1<<31);  /*mark processed*/

        bi=pQTxSts->e0 & 0xffff;  /*buffer index*/
        if(pP->d.idxQueTxDescTail!=bi) {  /*verify BI*/
            _PRTK_HWFAIL(("eth_cleanUpTx(): unmatching QTxSts[%d].BI:%d idxQTxDTail:%d\n",
                          idxSts,bi,pP->d.idxQueTxDescTail));
            /*@ recover*/
        }/*if*/


        /*statistics collection*/
        if(pQTxSts->e0 & (1<<30)) {  /*Sent without error*/
        //    pP->d.stats.tx_packets++;
        }else {  /*Tx failed due to error*/
            _PRTK_INFO(("eth_cleanUpTx(): Tx failed due to error QTxSts[%d]:0x%x\n",
                            idxSts, (unsigned int)pQTxSts->e0));
       //     pP->d.stats.tx_errors++;
            if(pQTxSts->e0&(1<<29)){  // pP->d.stats.tx_carrier_errors++;  /*loss of CRS*/
			}
            if(pQTxSts->e0&(1<<25)) {
       //         pP->d.stats.tx_fifo_errors++;     /*underrun*/
            }/*if*/
            if(pQTxSts->e0&(1<<24)){
				//  pP->d.stats.collisions++;  /*excessive collision*/
			}
        }/*else*/

        /*ahead Tx Descriptor Queue tail index*/
        pP->d.idxQueTxDescTail=IdxNext(pP->d.idxQueTxDescTail,LEN_QueTxDesc);
    }/*while*/

    return 0;
}/*eth_cleanUpTx()*/

/*****************************************************************************
* eth_isrTx()
*****************************************************************************/
static int  eth_isrTx(struct Ctlr *pD)
{
    _PRTK_ENTRY_ISR(("eth_isrTx(pD:0x%x)\n",(unsigned int)pD));

    eth_cleanUpTx(pD);
    eth_chkTxLvl(pD);  /*resume Tx if it was stopped*/
    return 0;    
}/*eth_isrTx()*/


/*****************************************************************************
* eth_restartTx()
*****************************************************************************/
static int  eth_restartTx(struct Ctlr *pD)
{
    struct ep9213Eth_info  *pP = &pD->priv;
    int  i;

    _PRTK_ENTRY(("eth_restartTx(pD:0x%p)\n",pD));

    /*disable int*/
    RegWr32(REG_GIntMsk, RegRd32(REG_GIntMsk)&~GIntMsk_IntEn);  /*turn off master INT control*/

    /*stop Tx and Tx DMA*/
    RegWr32(REG_TxCTL, RegRd32(REG_TxCTL)&~TxCTL_STxON);  /*Tx off*/
    RegWr32(REG_BMCtl, RegRd32(REG_BMCtl)|BMCtl_TxDis);  /*disable Tx DMA*/

    /*reset Tx DMA*/
    RegWr32(REG_BMCtl,BMCtl_TxChR|RegRd32(REG_BMCtl));  /*reset Tx DMA*/

    /*release Tx buffers*/
    for(i=0;i<LEN_QueTxDesc;i++) {
//        pP->d.stats.tx_dropped++;
    }/*for*/

    /*init Tx Queues and flush cache*/
    memset(pP->s.pQueTxSts, 0, sizeof(pP->s.pQueTxSts[0])*LEN_QueTxSts);  /*clear FP flag*/

    /*init variables*/
    pP->d.txStopped=0;
    pP->d.idxQueTxSts=pP->d.idxQueTxDescHead=pP->d.idxQueTxDescTail=0;

    /*init registers*/
    waitOnReg32(pD,REG_BMSts,BMCtl_TxChR,~BMCtl_TxChR,1);  /*wait to finish Tx DMA reset*/
    RegWr32(REG_TxSBA,pP->s.phyQueTxSts);  /*base addr of Tx Status Queue*/
    RegWr32(REG_TxSCA,pP->s.phyQueTxSts);  /*current addr*/
    RegWr16(REG_TxSBL,sizeof(pP->s.pQueTxSts[0])*LEN_QueTxSts);  /*base len*/
    RegWr16(REG_TxSCL,sizeof(pP->s.pQueTxSts[0])*LEN_QueTxSts);  /*current len*/
    RegWr32(REG_TxDBA,pP->s.phyQueTxDesc);  /*base addr of Tx Descriptor Queue*/
    RegWr32(REG_TxDCA,pP->s.phyQueTxDesc);  /*current addr*/
    RegWr16(REG_TxDBL,sizeof(pP->s.pQueTxDesc[0])*LEN_QueTxDesc);  /*base len*/
    RegWr16(REG_TxDCL,sizeof(pP->s.pQueTxDesc[0])*LEN_QueTxDesc);  /*current len*/

    /*start Tx and Tx DMA*/
    RegWr32(REG_TxCTL,RegRd32(REG_TxCTL)|TxCTL_STxON);  /*Tx on*/
    RegWr32(REG_BMCtl,RegRd32(REG_BMCtl)|BMCtl_TxEn);  /*enable Tx DMA*/

    /*enable int again*/
    RegWr32(REG_GIntMsk,RegRd32(REG_GIntMsk)|GIntMsk_IntEn);  /*turn on master INT control*/

    return 0;
}/*eth_restartTx()*/

/*****************************************************************************
* ep9213Eth_hardStartXmit()
*****************************************************************************/
static int  ep9213Eth_hardStartXmit(unsigned char *data, int len, Ctlr *pD) {

/*@swk  to check H/W defect of Tx Underrun Error caused by certain frame length*/
	extern int watchtrap, serialpost;
	static numsent = 0;
    struct ep9213Eth_info  *pP = &pD->priv;
    transmitDescriptor  *pQTxDesc;
    int                 idxQTxDescHd;
    int  filled;
//    int i;
//	unsigned int *pTemp;

    _PRTK_ENTRY(("ep9213Eth_hardStartXmit(data:%p,pD:%p)\n",data, pD));

    _PRTK_INFO(("ep9213Eth_hardStartXmit():len:0x%x\n",len));

    idxQTxDescHd = pP->d.idxQueTxDescHead;
    pQTxDesc = &pP->s.pQueTxDesc[idxQTxDescHd];

    /*check Tx Descriptor Queue fill-up level*/
    filled=idxQTxDescHd-pP->d.idxQueTxDescTail;
	_PRTK_INFO(("ep9213Eth_hardStartXmit(): filled is 0x%x\n", filled));
    if(filled<0)  filled+=LEN_QueTxDesc;
    filled += 1;

    if(/*!pP->d.txStopped&&*/LVL_TxStop <= filled) {  /*check Queue level*/
//        netif_stop_queue(pD);  /*no more Tx allowed*/
        pP->d.txStopped=1;
        _PRTK_INFO(("ep9213Eth_hardStartXmit(): Tx STOP requested, filled:%d\n",filled));
        if(LVL_TxStop<filled) {
            /*this situation can not be happen*/
            _PRTK_SYSFAIL(("ep9213Eth_hardStartXmit(): a Tx Request while stop\n"));
            return 1;
        }/*if*/
    }/*if*/

    /*fill up Tx Descriptor Queue entry*/
    if ( len < 60 )
    {
       pQTxDesc->e1 = 60;
	   memset(pP->s.pTxBufDesc[idxQTxDescHd].vaddr, 0, 60);
    }
    else
    {
       pQTxDesc->e1 = len;
    }
    pQTxDesc->e0 = PADDR(pP->s.pTxBufDesc[idxQTxDescHd].vaddr); 
    pQTxDesc->e1 |= idxQTxDescHd<<16;
//    pQTxDesc->f.af = 0;
    pQTxDesc->e1 |= 1<<31;
    _PRTK_INFO(("ep9213Eth_hardStartXmit(): Tx buf:%p len:%d QTxD[%d]:x%08x x%08lux\n",
               	data, len, (unsigned int)idxQTxDescHd,(unsigned int)pQTxDesc->e0, (unsigned long)pQTxDesc->e1));

	//copy data to Tx buffer
    memmove(pP->s.pTxBufDesc[idxQTxDescHd].vaddr, data, len);  
    memmove((char *)pP->s.pTxBufDesc[idxQTxDescHd].vaddr+6, pD->dev_addr, 6);
/*
						{ int i, j;
							iprint("XMIT\n");
							for(i = 0; i < len; i += 16){
								iprint("0x%x: ", i);
								for(j = 0; i + j < len; j ++)
									iprint("%02x ", data[i+j]);
								iprint("\n");
							}
						}
*/
	//Don't need this. Flush Tx buffer into memory
    //dma_cache_wback(pP->s.pTxBufDesc[idxQTxDescHd].vaddr, pQTxDesc->f.bl

    pP->d.idxQueTxDescHead=IdxNext(pP->d.idxQueTxDescHead,LEN_QueTxDesc);  /*ahead Tx Desc Queue*/
    _PRTK_INFO(("ep9213Eth_hardStartXmit(): pP->d.idxQueTxDescHead is 0x%x\n", 
			pP->d.idxQueTxDescHead));
//_PRTK_INFO(("NOT NOT NOT NOT enqueuing Tx descriptor\n"));
		numsent++;
	if (numsent > 1){
//		watchtrap++;
//		serialpost++;
	}
    RegWr32(REG_TxDEQ,1);  /*Enqueue a Tx Descriptor to the device*/
	_PRTK_INFO(("Current Da is 0%lux\n", RegRd32(REG_TxDCA)));
	post(0x70);
    return 0;
}/*ep9213Eth_hardStartXmit()*/

static void
ep93xxtransmit(Ether* edev)
{
	Block *bp;
	Ctlr *ctlr;
//	iprint("ep93xxtransmit, \n");
	ctlr = edev->ctlr;

	/*
	 * Called from both higher-level and interrupt routine.
	 * Clean any completed packets and try to fill the ring
	 * back up.
	 */
	ilock(&ctlr->tlock);

	while(bp = qget(edev->oq)){
			int i;
//		iprint("one hardstart done\n");
		/* this did not help. We're not getting MAC addresses. */
		memmove(bp->rp+6,edev->ea,sizeof(edev->ea));
//		for(i = 0; i < 6; i++)
//			iprint("EA BYTE %d is 0x%x\n", i+6,bp->rp[i+6]);
		ep9213Eth_hardStartXmit(bp->rp,BLEN(bp),ctlr);
		freeb(bp);
//		iprint("back from hardstart\n");
//		_dbg_ep9312Eth_dumpQueues(ctlr);
//		_dbg_phy_dumpReg(ctlr);
		
	}

	iunlock(&ctlr->tlock);
	_PRTK_INFO(("leaving xmit\n"));
}

static void
ep93xxreceive(Ether* edev)
{
	Ctlr *ctlr;

	ctlr = edev->ctlr;
	eth_isrRx(ctlr);
}

static void
ep93xxinterrupt(Ureg*, void* arg)
{
	Ctlr *ctlr;
	Ether *edev;
	//iprint("ep93xx int, edev %p, ctlr %p\n", edev, ctlr);
	edev = arg;
	ctlr = edev->ctlr;
	ep9213Eth_isr(ctlr);
//	iprint("ep93xx int done\n");
}


static int
ep93xxpnp(Ether* edev)
{
	Pcidev *p;
	Ctlr *ctlr;
	int i, id;

	SET(id, p, i);
	USED(id, p, i);

	/*
	 * If you only have one adaper and you know where it is,
	 * this is probably enough.
	 */
	iprint("ctlrhead %p\n", ctlrhead);
	if(ctlrhead == nil){

		ctlr = malloc(sizeof(Ctlr));


		/*
		 * This might be done here or in attach.
		 * Depends on the hardware.
		ep93xxmii(ctlr);
		 */

		ctlr->active = 1;
		ctlrhead = ctlr;
	}
	else
		return -1;
	ctlr->base_addr = (uchar *) MAC_BASE;
	iprint("done allocated, edev is %p, ctlr %p\n", edev, ctlr);
	edev->ctlr = ctlr;
	ctlr->edev = edev;
//	edev->port = 0;
	/*
	 * Don't need tbdf if no PCI.
	 * Set irq however your hardware is organised.
	 */
//	edev->irq = 0;

	edev->attach = ep93xxattach;
	edev->transmit = ep93xxtransmit;
	edev->interrupt = ep93xxinterrupt;
	edev->ifstat = ep93xxifstat;

	edev->arg = edev;
	edev->promiscuous = ep93xxpromiscuous;

	ep93xxlink(edev);

	return 0;
}

void
etherep93xxlink(void)
{
	poststr("etherlink\n");
	addethercard("ep93xx", ep93xxpnp);
}