/*
 *  Definitions for IO devices.  Used only in C.
 */

enum
{
	/* hardware counter frequency */
	ClockFreq=	508*1000,
};

/*
 *  IRQ's defined by 9302
 */
enum
{
	IRQCOMMRX=	2,
	IRQCOMMTX=	3,
	IRQTC1OI=	4,
	IRQTC2OI=	5,
	IRQAACINTR=	6,
	IRQDMAM2P0=	7,
	IRQDMAM2P1=	8,
	IRQDMAM2P2=	9,
	IRQDMAM2P3=	10,
	IRQDMAM2P4=	11,
	IRQDMAM2P5=	12,
	IRQDMAM2P6=	13,
	IRQDMAM2P7=	14,
	IRQDMAM2P8=	15,
	IRQDMAM2P9=	16,
	IRQDMAM2M0=	17,
	IRQDMAM2M1=	18,
	/* 19 reserved */
	IRQGPIOintr =	20,
	IRQGPI1intr=	21,
	IRQGPI2intr=	22,
	IRQUART1RX=	23,
	IRQUART1TX=	24,
	IRQUART2RX=	25,
	IRQUART2TX=	26,
	/* 27-31 reserved */
	IRQEXT0=	32,
	IRQEXT1=	33,
	/* 34 reserved */
	IRQ64HZ=	35,
	IRQdog=	36,
	IRQrtc=		37,

	IRQIRDA=	38,
	IRQENETMAC=39,
	IRQEXT2=	40,
	/* 41-44 reserved */
	IRQSSP1RX=	45,
	IRQSSP1TX=	46,
	/* 47-50 reserved */
	IRQTC3OI=	51,
	IRQUART1=	52,
	IRQSSP=		53,
	IRQUART2=	54,
	/* 55 reserved */
	IRQUSB=		56,
	/* 57-58 reserved */
	IRQGPIO=	59,
	IRQSA=		60
	/* 61-63 unused */
};

#define INTRMASK1 	(3 | 0x1f8 | 0x400 | 0x6000 | 0x780000 | 0x1000000 | 0xf8000000)
#define INTRMASK0	(0xFFFF | (7<<16) | (0x7f<<20))
/*
 *  GPIO lines (signal names from compaq document).  _i indicates input
 *  and _o output.
 * 
 * no use for now
enum
{
};
*/
/*
 *  types of interrupts
 */
enum
{
	GPIOrising,
	GPIOfalling,
	GPIOboth,
	IRQ,
};

/* hardware registers */
typedef struct Uartregs Uartregs;
struct Uartregs
{
	ulong	data;
	ulong	rxstatus;
	ulong	ctl[4];
	ulong	flags;
	ulong	interruptid;
	ulong	dmacontrol;
	ulong	modemcontrol;
	ulong	modemstatus;
	/* after this is hdlc, which I never intend to use */
};
extern Uartregs *uart3regs;
extern Uartregs *uart2regs;
extern Uartregs *uart1regs;

/* we have no use for GPIOs at present, and I have no 
 * idea if we ever will, so we leave these commented out for
 * now. I need to get this damned card working.
 * RGM
 */
#ifdef YEAHRIGHT

/* general purpose I/O lines control registers */
typedef struct GPIOregs GPIOregs;
struct GPIOregs
{
	ulong	level;		/* 1 == high */
	ulong	direction;	/* 1 == output */
	ulong	set;		/* a 1 sets the bit, 0 leaves it alone */
	ulong	clear;		/* a 1 clears the bit, 0 leaves it alone */
	ulong	rising;		/* rising edge detect enable */
	ulong	falling;	/* falling edge detect enable */
	ulong	edgestatus;	/* writing a 1 bit clears */
	ulong	altfunc;	/* turn on alternate function for any set bits */
};

extern GPIOregs *gpioregs;

/* extra general purpose I/O bits, output only */
enum
{
	EGPIO_prog_flash=	1<<0,
	EGPIO_pcmcia_reset=	1<<1,
	EGPIO_exppack_reset=	1<<2,
	EGPIO_codec_reset=	1<<3,
	EGPIO_exp_nvram_power=	1<<4,
	EGPIO_exp_full_power=	1<<5,
	EGPIO_lcd_3v=		1<<6,
	EGPIO_rs232_power=	1<<7,
	EGPIO_lcd_ic_power=	1<<8,
	EGPIO_ir_power=		1<<9,
	EGPIO_audio_power=	1<<10,
	EGPIO_audio_ic_power=	1<<11,
	EGPIO_audio_mute=	1<<12,
	EGPIO_fir=		1<<13,	/* not set is sir */
	EGPIO_lcd_5v=		1<<14,
	EGPIO_lcd_9v=		1<<15,
};
extern ulong *egpioreg;

/* Peripheral pin controller registers */
typedef struct PPCregs PPCregs;
struct PPCregs {
	ulong	direction;
	ulong	state;
	ulong	assignment;
	ulong	sleepdir;
	ulong	flags;
};
extern PPCregs *ppcregs;

/* Synchronous Serial Port controller registers */
typedef struct SSPregs SSPregs;
struct SSPregs {
	ulong	control0;
	ulong	control1;
	ulong	dummy0;
	ulong	data;
	ulong	dummy1;
	ulong	status;
};
extern SSPregs *sspregs;

/* Multimedia Communications Port controller registers */
typedef struct MCPregs MCPregs;
struct MCPregs {
	ulong	control0;
	ulong	reserved0;
	ulong	data0;
	ulong	data1;
	ulong	data2;
	ulong	reserved1;
	ulong	status;
	ulong	reserved[11];
	ulong	control1;
};
extern MCPregs *mcpregs;
#endif

/* unlike bitsy, it is dubious that we will ever need this */
#ifdef YEAHRIGHT
/*
 *  memory configuration
 */
enum
{
	/* bit shifts for pcmcia access time counters */
	MECR_io0=	0,
	MECR_attr0=	5,
	MECR_mem0=	10,
	MECR_fast0=	11,
	MECR_io1=	MECR_io0+16,
	MECR_attr1=	MECR_attr0+16,
	MECR_mem1=	MECR_mem0+16,
	MECR_fast1=	MECR_fast0+16,

	REFR_kapd=	29,
	REFR_eapd=	28,
	REFR_k1db2=	22,
	REFR_slfrsh=	31,
};

typedef struct MemConfRegs MemConfRegs;
struct MemConfRegs
{
	ulong	mdcnfg;		/* 0x00	dram */
	ulong	mdcas00;		/* 0x04	dram banks 0/1 */
	ulong	mdcas01;		/* 0x08 */
	ulong	mdcas02;		/* 0x0c */
	ulong	msc0;		/* 0x10	static */
	ulong	msc1;		/* 0x14 */
	ulong	mecr;		/* 0x18	pcmcia */
	ulong	mdrefr;		/* 0x1c	dram refresh */
	ulong	mdcas20;		/* 0x20	dram banks 2/3 */
	ulong	mdcas21;		/* 0x24 */
	ulong	mdcas22;		/* 0x28 */
	ulong	msc2;		/* 0x2c	static */
	ulong	smcnfg;		/* 0x30	SMROM config */
};
extern MemConfRegs *memconfregs;
#endif

/* is there any power management on this thing? */
#ifdef YEAHRIGHT

/*
 *  power management
 */

/* Power management ops */

typedef struct PowerRegs PowerRegs;
struct PowerRegs
{
	ulong	pmcr;	/* 0x0 Power manager control register */
	ulong	pssr;	/* 0x4 Power manager sleep status register */
	ulong	pspr;	/* 0x8 Power manager scratch pad register */
	ulong	pwer;	/* 0xc Power manager wakeup enable register */
	ulong	pcfr;	/* 0x10 Power manager general configuration register */
	ulong	ppcr;	/* 0x14 Power manager PPL configuration register */
	ulong	pgsr;	/* 0x18 Power manager GPIO sleep state register */
	ulong	posr;	/* 0x1c Power manager oscillator status register */
};
extern PowerRegs *powerregs;
#endif

typedef  struct OSTimerRegs OSTimerRegs;
struct OSTimerRegs
{
	ulong	load;
	ulong	value;
	ulong	control;
	ulong	clear;
};
extern OSTimerRegs* timer1, *timer2, *timer3;

typedef struct timer4regs	timer4regs;
struct timer4regs
{
	ulong	valuelow;
	ulong	enable_valuehigh;
};
extern timer4regs *timer4;

enum {
	TIMER4ENABLE = 0x100,
	OSTMRENABLE= 0x80,
	OSTMRPERIODIC=0x40,
	OSTMRFREERUN=0,
	OSTMR508KHZ = (1 << 3),
	OSTMR2KHZ   = (0 << 3)
};

typedef struct Intrregs Intrregs;
struct Intrregs
{
	ulong	icip;	/* pending IRQs */
	ulong	icfiq;	/* fiq status */
	ulong	icraw;	/* raw isr */
	ulong	icsel;	/* ip select */
	ulong	icenable;	/* enable */
	ulong	iclr;		/* clear enable */
	ulong 	icsir;	/* soft ints */
	ulong	icsirclear;	/* clear soft ints */
	ulong 	icprot;	/* protection enable */
	ulong	icvec;	/* vector address */
	ulong	icvecdef;	/* default icvec */
	ulong   pad1[0xd4/4];
	ulong	icvecaddress[16];	/* vectors */
	ulong   pad2[0xc0/4];
	ulong	icveccontrol[16]; /* control */
	ulong	pad3[0xda0/4];
	ulong	icppid[4];	/* peripheral ident register */
						/* LSbyte in [0], MSbyte in [3] */
	ulong	pad4[0xf000/4+4];
};
extern Intrregs *vic;



/*
 * Dont use this on bitsy drivers.
 * This is for compat with code compiled
 * from the ../pc directory
 */

typedef struct Pcidev	Pcidev;
typedef struct ISAConf	ISAConf;
struct Pcidev { int dummy; };
struct ISAConf{ int port; int irq; };
int pcmspecial(char*, ISAConf*);
