#include	"u.h"
#include	"../port/lib.h"
#include	"mem.h"
#include	"dat.h"
#include	"fns.h"
#include	"io.h"
#include	"ureg.h"
#include	"../port/error.h"

/*
 *  to avoid mmu and cash flushing, we use the pid register in the MMU
 *  to map all user addresses.  Although there are 64 possible pids, we
 *  can only use 31 because there are only 32 protection domains and we
 *  need one for the kernel.  Pid i is thus associated with domain i.
 *  Domain 0 is used for the kernel.
 */

/* real protection bits */
enum
{
	/* level 1 descriptor bits */
	L1TypeMask=	(3<<0),
	L1Invalid=	(0<<0),
	L1PageTable=	(1<<0),
	L1Section=	(2<<0),
	L1Cached=	(1<<3),
	L1Buffered=	(1<<2),
	L1DomShift=	5,
	L1Domain0=	(0<<L1DomShift),
	L1KernelRO=	(0x0<<10),
	L1KernelRW=	(0x1<<10),
	L1UserRO=	(0x2<<10),
	L1UserRW=	(0x3<<10),
	L1SectBaseMask=	(0xFFF<<20),
	L1PTBaseMask=	(0x3FFFFF<<10),
	
	/* level 2 descriptor bits */
	L2TypeMask=	(3<<0),
	L2SmallPage=	(2<<0),
	L2LargePage=	(1<<0),
	L2Cached=	(1<<3),
	L2Buffered=	(1<<2),
	L2KernelRW=	(0x55<<4),
	L2UserRO=	(0xAA<<4),
	L2UserRW=	(0xFF<<4),
	L2PageBaseMask=	(0xFFFFF<<12),

	/* domain values */
	Dnoaccess=	0,
	Dclient=	1,
	Dmanager=	3,
};

ulong *l1table;
static int mmuinited;

/*
 *  We map all of memory, flash, and the zeros area with sections.
 *  Special use space is mapped on the fly with regmap.
 */

char *loadformat = 
"\tMOVW\t$0x%x, R3\n\tMOVW\t$0x%ulx, R2\n\tMOVW\tR2,0(R3)\n";
void
main(void)
{
	ulong a, o;
	ulong *t;
	int i;
	unsigned long pteval;
	unsigned long *l1base = (unsigned long *) 0x4000;
	unsigned long *l2base = (unsigned long *) 0x7000;
	for(i = 0; i < 32; i++)
		*l2base++ = *l1base++;

	
#ifdef no
	/* let's try a loop ... */
	for(i = 0, l1base = 0x7000; i < 128; i++) {
		*l1base++ = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((i<<18)&L1SectBaseMask);
	}
	for(i = 0, l1base=0x4000; i < 128; i++) {
		*l1base++ = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((i<<18)&L1SectBaseMask);
	}

#endif
#define YA
#ifdef YA
	/* low memory */
	print("/* map low memory, first 8M*/\n");
	for(i = 0; i < 8*4; i+= 4) {
		pteval = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((i<<18)&L1SectBaseMask);

		print(loadformat, i+l1base, pteval);
	}

	/* high memory -- l1base is 0x7000 */
	print("/*Map first 8M of 0xc0000000*/\n");
	l1base = (ulong*)0x8000 + 0x3000/4;
	for(i = 0; i < 8*4; i += 4) {
		pteval = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((i<<18)&L1SectBaseMask);

		print(loadformat, i+l1base, pteval);
	}


	/* ok, do the POST and UART regions */
	print("/* map PCI I/O regions for POST code */\n");
		l1base = (ulong*)0x8000;
		i = 0x11800000>>18;
		pteval = L1Section | L1KernelRW| L1Domain0 
			| ((0x11800000)&L1SectBaseMask);
	
		print(loadformat, i+l1base, pteval);
	print("/*map region containing UARTS etc. */\n");
		i = (unsigned long)0x80800000>>18;
		pteval = L1Section | L1KernelRW| L1Domain0 
			| ((0x80800000)&L1SectBaseMask);
	
		print(loadformat, i+l1base, pteval);
#endif
#ifdef YA

	/* get a prototype level 1 page */
	l1table = (ulong *)0x4000;
//	memset(l1table, 0, 16*1024);
	*l1table = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((0<<20)&L1SectBaseMask);
	l1table++;
	*l1table = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((1<<20)&L1SectBaseMask);
	l1table++;
	*l1table = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((2<<20)&L1SectBaseMask);
	l1table++;
	*l1table = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((3<<20)&L1SectBaseMask);
	l1table++;
	*l1table = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((4<<20)&L1SectBaseMask);
	l1table++;
	*l1table = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((5<<20)&L1SectBaseMask);
	l1table++;
	*l1table = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((6<<20)&L1SectBaseMask);
	l1table++;
	*l1table = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((7<<20)&L1SectBaseMask);
	l1table++;

/* now map the high memory. This is going to be at 
 * the fourth page, i.e. 16384 + 3*4096
 */
	l1table = (ulong *) 0x7000;
	*l1table = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((0<<20)&L1SectBaseMask);
	l1table++;
	*l1table = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((1<<20)&L1SectBaseMask);
	l1table++;
	*l1table = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((2<<20)&L1SectBaseMask);
	l1table++;
	*l1table = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((3<<20)&L1SectBaseMask);
	l1table++;
	*l1table = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((4<<20)&L1SectBaseMask);
	l1table++;
	*l1table = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((5<<20)&L1SectBaseMask);
	l1table++;
	*l1table = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((6<<20)&L1SectBaseMask);
	l1table++;
	*l1table = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((7<<20)&L1SectBaseMask);
	l1table++;
	/* now for PC/104 I/O so we have post ...
 	*/
	/* >> 18 not 20 as entries are 4 bytes */
	l1table = (ulong *)((0x11800000>>18) + 0x4000);
	*l1table = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((0x11800000)&L1SectBaseMask);
	l1table = (ulong *)(((unsigned long)0x80800000>>18) + 0x4000);
	*l1table = L1Section | L1KernelRW| L1Domain0 
			| L1Cached | L1Buffered
			| ((0x80800000)&L1SectBaseMask);
#endif
}
