#include	"u.h"
#include	"../port/lib.h"
#include	"mem.h"
#include	"dat.h"
#include	"fns.h"
#include	"io.h"
#include	"../port/error.h"

#include	"../port/netif.h"

void ledpost(int);

enum
{
	/* ctl[0] bits */
	Parity=		1<<1,
	Even=		1<<2,
	Stop2=		1<<3,
	FifoEnable=	1<<4,
	Bits8=		3<<5,
	Bits7=		2<<5,
	Brk=		1<<0,

	/* ctl[3] bits */
	LBE=		1 << 7,
	RTIE=		1 << 6,
	TIE=		1 << 5,
	RIE=		1 << 4,
	MSIE=		1 << 3,
	UARTE = 	1 << 0,


	/* flag bits */
	TXFE=		1<<7,
	RXFF=		1<<6,
	TXFF=		1<<5,
	RXFE=		1<<4,
	BUSY=		1<<3,
	DCD=		1<<2,
	DSR=		1<<1,
	CTS=		1<<0,

	/* status bits */
	BreakError =		1<<2,	/* transmitting */
	ParityError=	1<<1,
	FrameError=	1<<0,
	Overrun=		1<<3,
};

extern PhysUart sa930xphysuart;

//static
Uart sa930xuart[1] = {
{	.regs		= UART1REGS,
	.name	= "serialport1",
	.freq		= ClockFreq,
	.bits		= 8,
	.stop		= 1,
	.parity	= 'n',
	.baud	= 115200,
	.phys	= &sa930xphysuart,
	.special	= 0,
	.next		= 0, },

};

#define R(p) ((Uartregs*)((p)->regs))
#define SR(p) ((Uartregs*)((p)->saveregs))

/*
 *  enable a port's interrupts.  set DTR and RTS
 */
static void
sa930x_uartenable(Uart *p, int intena)
{
	ulong s;
	/*
	post(0xf);
	poststr("uart enable. R(p) is ");
	put32((unsigned long ) R(p)); poststr("\n");
	poststr("try to touch it ...");
	post(0x10);
	put32(*(unsigned long *)R(p));poststr("\n");
	poststr("INTS ON UART ARE "); put32(intena); poststr("\n");
	*/
	post(0x11);
	s = R(p)->ctl[3] & ~(RTIE|TIE|UARTE);
	if(intena)
		R(p)->ctl[3] = s |RTIE|/*TIE|*/UARTE;
	else
		R(p)->ctl[3] = s | UARTE;
	post(0x12);
}

/*
 *  disable interrupts. clear DTR, and RTS
 */
static void
sa930x_uartdisable(Uart *p)
{
	R(p)->ctl[3] &= ~(RTIE|TIE|UARTE);
}

static long
sa930x_uartstatus(Uart *p, void *buf, long n, long offset)
{
	char str[256];
	ulong ctl0;

	ctl0 = R(p)->ctl[0];
	snprint(str, sizeof(str),
		"b%d c%d d%d e%d l%d m%d p%c r%d s%d i%d\n"
		"dev(%d) type(%d) framing(%d) overruns(%d)%s%s%s%s\n",

		p->baud,
		p->hup_dcd, 
		0,
		p->hup_dsr,
		(ctl0 & Bits8) == 3 ? 8 : 7,
		0, 
		(ctl0 & Parity) ? ((ctl0 & Even) ? 'e' : 'o') : 'n',
		0,
		(ctl0 & Stop2) ? 2 : 1,
		1,

		p->dev,
		p->type,
		p->ferr,
		p->oerr, 
		"",
		"",
		"",
		"" );
	return readstr(offset, buf, n, str);
}

/*
 *  set the buad rate
 */
static int
sa930x_uartbaud(Uart *p, int rate)
{
	ulong brconst;
	ulong ctl3;

	if(rate <= 0)
		return -1;

	/* disable */
	ctl3 = R(p)->ctl[3];
	R(p)->ctl[3] = 0;

	brconst = p->freq/(16*rate) - 1;
	R(p)->ctl[1] = (brconst>>8) & 0xf;
	R(p)->ctl[2] = brconst & 0xff;

	/* reenable */
	R(p)->ctl[3] = ctl3;

	p->baud = rate;
	return 0;
}

/*
 *  send a break
 */
static void
sa930x_uartbreak(Uart *p, int ms)
{
	if(ms == 0)
		ms = 200;

	R(p)->ctl[1] |= Brk;
	tsleep(&up->sleep, return0, 0, ms);
	R(p)->ctl[1] &= ~Brk;
}

/*
 *  set bits/char
 */
static int
sa930x_uartbits(Uart *p, int n)
{
	ulong ctl0, ctl3;

	ctl0 = R(p)->ctl[0];
	switch(n){
	case 7:
		ctl0 &= ~Bits8;
		ctl0 |= Bits7;
		break;
	case 8:
		ctl0 |= Bits8;
		break;
	default:
		return -1;
	}

	/* disable */
	ctl3 = R(p)->ctl[3];
	R(p)->ctl[3] = 0;

	R(p)->ctl[0] = ctl0;

	/* reenable */
	R(p)->ctl[3] = ctl3;

	p->bits = n;
	return 0;
}

/*
 *  set stop bits
 */
static int
sa930x_uartstop(Uart *p, int n)
{
	ulong ctl0, ctl3;

	ctl0 = R(p)->ctl[0];
	switch(n){
	case 1:
		ctl0 &= ~Stop2;
		break;
	case 2:
		ctl0 |= Stop2;
		break;
	default:
		return -1;
	}

	/* disable */
	ctl3 = R(p)->ctl[3];
	R(p)->ctl[3] = 0;

	R(p)->ctl[0] = ctl0;

	/* reenable */
	R(p)->ctl[3] = ctl3;

	p->stop = n;
	return 0;
}

/*
 *  turn on/off rts
 */
static void
sa930x_uartrts(Uart*, int)
{
}

/*
 *  turn on/off dtr
 */
static void
sa930x_uartdtr(Uart*, int)
{
}

/*
 *  turn on/off modem flow control on/off (rts/cts)
 */
static void
sa930x_uartmodemctl(Uart *p, int on)
{
	if(on) {
	} else {
		p->cts = 1;
	}
}

/*
 *  set parity
 */
static int
sa930x_uartparity(Uart *p, int type)
{
	ulong ctl0, ctl3;

	ctl0 = R(p)->ctl[0];
	/* bug, they did not clear the bits out */
	ctl0 &= ~(Parity|Even);
	switch(type){
	case 'e':
		ctl0 |= Parity|Even;
		break;
	case 'o':
		ctl0 |= Parity;
		break;
	default:
		ctl0 &= ~(Parity|Even);
		break;
	}

	/* disable */
	ctl3 = R(p)->ctl[3];
	R(p)->ctl[3] = 0;

	R(p)->ctl[0] = ctl0;

	/* reenable */
	R(p)->ctl[3] = ctl3;

	return 0;
}

/*
 *  restart output if not blocked and OK to send
 */
static void
sa930x_uartkick(Uart *p)
{
	int i;

//	R(p)->ctl[3] &= ~TIE;

	if(p->cts == 0 || p->blocked)
		return;

	for(i = 0; i < 1024; i++){
		if((R(p)->flags & TXFF)){
			//R(p)->ctl[3] |= TIE;
			break;
		}
		if(p->op >= p->oe && uartstageoutput(p) == 0)
			break;
		R(p)->data = *p->op++;
	}
}

/*
 *  take an interrupt
 */
void
sa930x_uartintr(Ureg*, void *x)
{
	Uart *p;
	ulong s;
	ulong stat;
	Uartregs *regs;
//	poststr("U");
	p = x;
	regs = p->regs;
//	put32(regs->interruptid);poststr("\n");
//	regs->interruptid = 0x1f;
//	put32(regs->flags);poststr("\n");
	/* receiver interrupt, snarf bytes */
//	poststr("flags & RXF "); put32(regs->flags & RXFE); poststr("\n");
	while(!(regs->flags & RXFE))
		uartrecv(p, regs->data);

	/* remember and reset interrupt causes */
	s = regs->flags;
//	regs->flags |= s;

/* just kick it */
	/* if it is not full, then kick it. */
	if(!(s & TXFF)){
		/* transmitter interrupt, restart */
		uartkick(p);
	}
	stat = regs->rxstatus;
	if(stat & (ParityError|FrameError|Overrun)){
		if(stat & ParityError)
			p->parity++;
		if(stat & FrameError)
			p->ferr++;
		if(stat & Overrun)
			p->oerr++;
	}

	/* receiver interrupt, snarf bytes until empty */
//	poststr("flags & RXF "); put32(regs->flags & RXFE); poststr("\n");
	while(!(regs->flags & RXFE))
		uartrecv(p, regs->data);
}

static Uart*
sa930x_pnp(void)
{
	return sa930xuart;
}

static int
sa930x_getc(Uart *uart)
{
	Uartregs *ur;

	ur = uart->regs;
	while((ur->flags & RXFE))
		;
	return ur->data;
}

void
debugputc(unsigned long c) {

	volatile Uartregs *ur = (Uartregs *) UART1REGS;
	static int firsttime = 0;
	static unsigned long x = '0';
	volatile unsigned long *ld = (unsigned long *) 0x808c0000;
	volatile unsigned long *ls = (unsigned long *) 0x808c0018;
//	*(volatile unsigned long *) UART1REGS = (unsigned long) '3';
	/* clean it up and be paranoid about it */
	c &= 0x7f;

	ledpost(c);	
	ledpost(0x20);
	if (firsttime == 0xabcd) {
		ledpost(0x21);
		ur->ctl[0] = 0x70;
		ur->ctl[1] = 0;
		ur->ctl[2] = 0x70;
		ur->ctl[3] = 1;
		ur->data = 'F';
		ledpost(0x22);
		firsttime = 0;
	}

	/* wait for output ready */
	while(*ls & TXFF)
		ledpost(0x23);

	*(unsigned long *) UART1REGS = c;
	
	ledpost(0x24);
	while((*ls & BUSY))
		ledpost(0x25);
	ledpost(0x2f);
}
static void
sa930x_putc(Uart *uart, int c)
{
	Uartregs *ur;

	ur = uart->regs;
	/* wait for output ready */
	while(ur->flags & TXFF)
		;
	ur->data = c;
	while((ur->flags & BUSY))
		;
}

PhysUart sa930xphysuart = {
	.name=		"sa930x",
	.pnp= 		sa930x_pnp,
	.enable=		sa930x_uartenable,
	.disable=		sa930x_uartdisable,
	.bits=		sa930x_uartbits,
	.kick=		sa930x_uartkick,
	.modemctl=	sa930x_uartmodemctl,
	.baud=		sa930x_uartbaud,
	.stop=		sa930x_uartstop,
	.parity=		sa930x_uartparity,
	.dobreak=		sa930x_uartbreak,
	.rts=			sa930x_uartrts,
	.dtr=			sa930x_uartdtr,
	.status=		sa930x_uartstatus,
	.power=		0,
	.getc=		sa930x_getc,
	.putc=		sa930x_putc,
};

/*
 *  setup all uarts (called early by main() to allow debugging output to
 *  a serial port)
 */
void
sa930x_uartsetup(void)
{
	Uart *p;
	post(0xff);
	/* external serial port (eia0) */
	p = &sa930xuart[0];
	p->regs = (void *) UART1REGS; //mapspecial(UART1REGS, sizeof(Uartregs));
//	put32((unsigned long)p->regs); poststr(" is regs\n");
	p->saveregs = xalloc(sizeof(Uartregs));
	/* set eia0 up as a console */
//	poststr("call uartctl\n");
//	uartctl(p, "b115200 l8 pn s1");
	uartctl(p, "");
//	poststr("done uartctl\n");

	(*p->phys->enable)(p, 1);
//	poststr("done enable\n");
	p->console = 1;
	consuart = p;
	post(0xf8);
	intrenable(IRQ, IRQUART1RX, sa930x_uartintr, p, p->name);
	post(0xf7);
	intrenable(IRQ, IRQUART1TX, sa930x_uartintr, p, p->name);
	intrenable(IRQ, IRQUART1, sa930x_uartintr, p, p->name);
	post(0xfe);

}

static  void
uartcpy(Uartregs *to, Uartregs *from)
{
	to->ctl[0] = from->ctl[0];
//	to->ctl[1] = from->ctl[1];
//	to->ctl[2] = from->ctl[2];
	to->ctl[3] = from->ctl[3];

}
